import os
from unittest import mock

import pytest
from click import NoSuchOption
from click import UsageError

from coworks.cws.client import client


class TestClass:

    def test_wrong_cmd(self, example_dir):
        with pytest.raises(UsageError) as pytest_wrapped_e:
            client.main(['--project-dir', '.', 'cmd'], 'cws', standalone_mode=False)
        assert pytest_wrapped_e.type == UsageError
        assert pytest_wrapped_e.value.args[0] == "No such command 'cmd'."

    @mock.patch.dict(os.environ, {"FLASK_APP": "command:app"})
    def test_cmd(self, example_dir, capsys):
        client.main(['--project-dir', '.', 'test'], 'cws', standalone_mode=False)
        captured = capsys.readouterr()
        assert captured.out == "test command with a=default/test command with b=value"

    def test_cmd_wrong_option(self, example_dir):
        with pytest.raises(NoSuchOption) as pytest_wrapped_e:
            client.main(['--project-dir', '.', 'test', '-t', 'wrong'], 'cws', standalone_mode=False)
        assert pytest_wrapped_e.type == NoSuchOption
        assert pytest_wrapped_e.value.args[0] == "No such option: -t"

    @mock.patch.dict(os.environ, {"FLASK_APP": "command:app"})
    def test_cmd_right_option(self, example_dir, capsys):
        client.main(['--project-dir', '.', 'test', '-a', 'right'], 'cws', standalone_mode=False)
        captured = capsys.readouterr()
        assert captured.out == "test command with a=right/test command with b=value"

    @mock.patch.dict(os.environ, {"FLASK_APP": "command:app"})
    def test_cmd_wrong_b_option(self, example_dir, capsys):
        with pytest.raises(NoSuchOption) as pytest_wrapped_e:
            client.main(['--project-dir', '.', 'test', '-b', 'right'], 'cws', standalone_mode=False)
        assert pytest_wrapped_e.type == NoSuchOption

    @mock.patch.dict(os.environ, {"FLASK_APP": "command:app"})
    def test_cmd_right_b_option(self, example_dir, capsys):
        try:
            client.main(['--project-dir', '.', 'test', '--b', 'right'], 'cws', standalone_mode=False)
        finally:
            os.unsetenv("FLASK_RUN_FROM_CLI")
        captured = capsys.readouterr()
        assert captured.out == "test command with a=default/test command with b=right"

    @mock.patch.dict(os.environ, {"FLASK_APP": "command:app"})
    def test_v1_cmd(self, example_dir, capsys):
        client.main(['--project-dir', '.', '-s', 'v1', 'test', '-a', 'right'], 'cws',
                    standalone_mode=False)
        captured = capsys.readouterr()
        assert captured.out == "test command v1 with a=right/test command v1 with b=value1"



from coworks import Blueprint
from coworks import entry


class TestBlueprint(Blueprint):
    """Test blueprint.
    This blueprint proposes a template blueprint for CI/CD purpose.

    It has a single get entry for testing purpose.

    .. versionchanged:: 0.7.3
        Added.
    """

    def __init__(self, name='test', **kwargs):
        super().__init__(name=name, **kwargs)

    @property
    def test_workspaces(self):
        return ['dev']

    @entry(stage='dev')
    def post_reset(self):
        """Entry to reset the test environment."""
        return 'ok'

    @entry(stage='dev')
    def get(self):
        """Test entry.

        Returns the next entry test case if needed. {'path':..., 'expected value':...}
        """
        self.post_reset()
        self.test()
        self.post_reset()

    def test(self):
        return 'ok'



import os
from unittest import mock
from unittest.mock import Mock

from flask import request
from flask import url_for
from werkzeug.exceptions import Forbidden
from werkzeug.exceptions import MethodNotAllowed
from werkzeug.exceptions import NotFound
from werkzeug.exceptions import UnprocessableEntity

from coworks import TechMicroService
from coworks import entry
from ..event import get_event
from ..ms import GlobalMS
from ..ms import SimpleMS


class CookiesMS(TechMicroService):

    @entry
    def get_cookies(self):
        return request.cookies.to_dict()

    @entry
    def get_url_for(self, external=None):
        return url_for('get_cookies', _external=external)


class ErrorMS(TechMicroService):

    def __init__(self):
        super().__init__()
        self.err = Mock()

        @self.errorhandler(Forbidden.code)
        def handle(e):
            self.err(e)
            return "ok"

    @entry
    def get(self):
        return "Error 404", 404

    @entry
    def get_exception(self):
        return int("a")

    @entry
    def get_http_exception(self):
        raise Forbidden()


@mock.patch.dict(os.environ, {"TOKEN": "token"})
class TestClass:
    def test_request_arg(self, empty_aws_context):
        app = SimpleMS()
        with app.app_context() as c:
            response = app(get_event('/', 'get'), empty_aws_context)
            assert response['statusCode'] == 200
            assert response['body'] == "get"
            assert 'Content-Type' in response['headers']
            assert response['headers']['Content-Type'] == 'text/plain; charset=utf-8'
            assert 'Content-Length' in response['headers']
            assert response['headers']['Content-Length'] == str(len(response['body']))
            response = app(get_event('/', 'post'), empty_aws_context)
            assert response['statusCode'] == MethodNotAllowed.code
            response = app(get_event('/get1', 'get'), empty_aws_context)
            assert response['statusCode'] == NotFound.code
            response = app(get_event('/content', 'get'), empty_aws_context)
            assert response['statusCode'] == 200
            assert response['body'] == "get content"
            response = app(
                get_event('/content/{value}', 'get', entry_path_parameters={'value': 3}),
                empty_aws_context)
            assert response['statusCode'] == 200
            assert response['body'] == "get content with 3"
            response = app(
                get_event('/content/{value}/{other}', 'get', entry_path_parameters={'value': 3, 'other': 'other'}),
                empty_aws_context)
            assert response['statusCode'] == 200
            assert response['body'] == "get content with 3 and other"
            response = app(
                get_event('/content', 'post', body={"other": 'other'}),
                empty_aws_context)
            assert response['statusCode'] == 200
            assert response['body'] == "post content without value but other"
            response = app(
                get_event('/content/{value}', 'post', entry_path_parameters={'value': 3}, body={"other": 'other'}),
                empty_aws_context)
            assert response['statusCode'] == 200
            assert response['body'] == "post content with 3 and other"
            response = app(get_event(
                '/content/{value}', 'post', entry_path_parameters={'value': 3}, body="other"),
                empty_aws_context)
            assert response['statusCode'] == 200
            assert response['body'] == "post content with 3 and other"
            response = app(
                get_event('/content/{value}', 'post', entry_path_parameters={'value': 3},
                          body={"other": 'other', "value": 5}),
                empty_aws_context)
            assert response['statusCode'] == UnprocessableEntity.code

    def test_request_kwargs(self, empty_aws_context):
        app = SimpleMS()
        with app.app_context() as c:
            response = app(get_event('/kwparam1', 'get', params={'value': [5]}), empty_aws_context)
            assert response['statusCode'] == 200
            assert response['body'] == "get **param with only 5"
            response = app(get_event('/kwparam1', 'get', params={"other": ['other'], "value": [5]}), empty_aws_context)
            assert response['statusCode'] == 422
            response = app(get_event('/kwparam1', 'get', params={"value": [5]}), empty_aws_context)
            assert response['statusCode'] == 200
            assert response['body'] == "get **param with only 5"
            response = app(get_event('/kwparam1', 'get', body={"other": 'other', "value": 5}), empty_aws_context)
            assert response['statusCode'] == 200
            assert response['body'] == "get **param with only 0"
            response = app(get_event('/kwparam2', 'get', params={"other": ['other'], "value": [5]}), empty_aws_context)
            assert response['statusCode'] == 200
            assert response['body'] == "get **param with 5 and ['other']"
            response = app(get_event('/kwparam2', 'get', body={"other": 'other', "value": 5}), empty_aws_context)
            assert response['statusCode'] == 200
            assert response['body'] == "get **param with 0 and []"
            response = app(get_event('/kwparam2', 'put', body={"other": 'other', "value": 5}), empty_aws_context)
            assert response['statusCode'] == 200
            assert response['body'] == "get **param with 5 and ['other']"
            response = app(get_event('/kwparam2', 'put', params={"other": ['other'], "value": [5]}), empty_aws_context)
            assert response['statusCode'] == 200
            assert response['body'] == "get **param with 0 and []"
            response = app(get_event('/extended/content', 'get'), empty_aws_context)
            assert response['statusCode'] == 200
            assert response['body'] == "hello world"
            event = get_event('/dir/content', 'get')
            del event['headers']['content-type']
            response = app(event, empty_aws_context)
            assert response['statusCode'] == 200
            assert response['body'] == {'msg': 'hello world'}

    def test_request_globals(self, empty_aws_context):
        app = GlobalMS()
        with app.app_context() as c:
            response = app(get_event('/event/method', 'get'), empty_aws_context)
            assert response['statusCode'] == 200
            assert response['body'] == "GET"

    def test_request_cookies(self, empty_aws_context):
        app = CookiesMS()
        with app.app_context() as c:
            response = app(get_event('/cookies', 'get'), empty_aws_context)
            assert response['statusCode'] == 200
            assert response['body'] == {'session': '70b58773-af3e-4153-a5ab-5356481ea87e'}

    def test_request_no_content_type(self, empty_aws_context):
        """If content-type is not defined then default is application/json"""
        app = SimpleMS()
        with app.app_context() as c:
            event = get_event('/content', 'post', body={'other': 'other'})
            del event['headers']['content-type']
            response = app(event, empty_aws_context)
            assert response['statusCode'] == 200
            assert response['body'] == "post content without value but other"

    def test_request_error(self, empty_aws_context):
        app = ErrorMS()
        with app.app_context() as c:
            event = get_event('/', 'get')
            response = app(event, empty_aws_context)
            assert response['statusCode'] == NotFound.code
            assert response['body'] == "Error 404"
            event = get_event('/exception', 'get')
            response = app(event, empty_aws_context)
            assert response['statusCode'] == 500
            app.err.assert_not_called()
            event = get_event('/http/exception', 'get')
            response = app(event, empty_aws_context)
            assert response['statusCode'] == 200
            app.err.assert_called_once()

    def test_url_for(self, empty_aws_context):
        app = CookiesMS()
        with app.app_context() as c:
            event = get_event('/url/for', 'get')
            response = app(event, empty_aws_context)
            assert response['statusCode'] == 200
            assert response['body'] == '/dev/cookies'
            event = get_event('/url/for', 'get', params={'external': True})
            response = app(event, empty_aws_context)
            assert response['statusCode'] == 200
            assert response['body'] == 'https://htzd2rneg1.execute-api.eu-west-1.amazonaws.com/dev/cookies'
            event = get_event('/url/for', 'get', params={'external': False})
            response = app(event, empty_aws_context)
            assert response['statusCode'] == 200
            assert response['body'] == '/dev/cookies'



# noinspection PyPep8Naming
from airflow import DAG as BizMicroService
from airflow.decorators import dag

from coworks.biz.group import NeoRezoServiceAsyncGroup
from coworks.biz.group import TechMicroServiceAsyncGroup
from coworks.biz.operators import AsyncTechServicePullOperator
from coworks.biz.operators import BranchTechMicroServiceOperator
from coworks.biz.operators import NeoRezoServiceOperator
from coworks.biz.operators import TechMicroServiceOperator
from coworks.biz.sensors import AsyncTechMicroServiceSensor
from coworks.biz.sensors import TechMicroServiceSensor


def biz(*biz_args, **biz_kwargs):
    """dag decorator redefined to create the tech microservice for triggering it in the future."""

    if 'doc_md' in biz_kwargs:
        biz_kwargs['doc_md'] += """
<img style="margin-bottom:auto;width:100px;" src="https://neorezo.io/assets/img/logo_neorezo.png"/>
"""

    def wrapper(func):
        return dag(*biz_args, **biz_kwargs)(func)

    return wrapper


_all__ = (
    AsyncTechServicePullOperator, BranchTechMicroServiceOperator, TechMicroServiceOperator,
    AsyncTechMicroServiceSensor, TechMicroServiceSensor,
    TechMicroServiceAsyncGroup,
    NeoRezoServiceAsyncGroup, NeoRezoServiceOperator,
    BizMicroService, biz,
)



import logging
import os
import traceback
import typing as t
from functools import partial
from functools import update_wrapper

from aws_xray_sdk import global_sdk_config
# from aws_xray_sdk.core import patch_all
from aws_xray_sdk.core.exceptions.exceptions import SegmentNotFoundException
from aws_xray_sdk.core.recorder import TRACING_NAME_KEY
from aws_xray_sdk.ext.flask.middleware import XRayMiddleware

from coworks.globals import request
from coworks.wrappers import CoworksResponse

if t.TYPE_CHECKING:
    from coworks import TechMicroService
    from aws_xray_sdk.core import AWSXRayRecorder

LAMBDA_NAMESPACE = 'lambda'
REQUEST_NAMESPACE = 'flask'
COWORKS_NAMESPACE = 'coworks'


class XRay:
    def __init__(self, app: "TechMicroService", recorder: "AWSXRayRecorder", name="xray"):
        self._app = app
        self._recorder = recorder
        self._name = name

        if app is not None:
            self.init_app(app)

        # Bug in aws_xray_sdk : service must be defined or exception will be raised in execution if no dynamic_naming
        if not recorder.service:
            recorder.configure(service=COWORKS_NAMESPACE)

    def init_app(self, app):

        # Not available in flask client
        if os.getenv("FLASK_RUN_FROM_CLI"):
            return

        XRayMiddleware(app, self._recorder)
        app.logger.debug(f"Initializing xray extension {self._name}")

        if app.debug:
            logging.getLogger('aws_xray_sdk').setLevel(logging.DEBUG)

        if global_sdk_config.sdk_enabled():
            # Checks XRay is available
            try:
                self._recorder.current_segment()
            except SegmentNotFoundException:
                pass
            else:
                # Captures routes
                # patch_all() # Bug in aws_xray if engine is defined by flask_sqlalchemy
                app.errorhandler(500)(self.capture_exception)
                self.capture_routes()
                return

        self._app.logger.debug("Skipped capture routes because the SDK is currently disabled.")

    @staticmethod
    def capture(recorder):
        # Decorator to trace function calls on XRay.
        def xray_decorator(function):
            def function_captured(*args, **kwargs):
                subsegment = recorder.current_subsegment()
                if subsegment:
                    call_data = {
                        'args': args[1:],
                        'kwargs': kwargs,
                    }
                    subsegment.put_metadata(function.__name__, call_data, COWORKS_NAMESPACE)

                response = function(*args, **kwargs)

                if subsegment:
                    subsegment.put_metadata(f'{function.__name__}.response', response, COWORKS_NAMESPACE)

                return response

            if global_sdk_config.sdk_enabled():
                # Checks XRay is available
                try:
                    recorder.current_segment()
                except SegmentNotFoundException:
                    pass
                else:
                    # Captures function
                    wrapped_fun = update_wrapper(function_captured, function)
                    return recorder.capture(name=function.__name__)(wrapped_fun)

            return function

        return xray_decorator if os.getenv(TRACING_NAME_KEY) else lambda x: x

    def capture_routes(self):
        for rule in self._app.url_map.iter_rules():
            view_function = self._app.view_functions[rule.endpoint]

            def route_captured(_view_function, *args, **kwargs):
                aws_context = request.aws_context
                subsegment = self._recorder.current_subsegment()

                if subsegment:
                    try:
                        subsegment.put_metadata('context', lambda_context_to_json(aws_context), LAMBDA_NAMESPACE)
                        metadata = {
                            'service': self._app.name,
                            'environ': request_environ_to_json(request),
                        }
                        if request.is_json:
                            try:
                                metadata['json'] = request.json
                            except (Exception,):
                                metadata['data'] = request.get_data(cache=False, as_text=True)
                        elif request.is_multipart:
                            metadata['multipart'] = request.form.to_dict(False)
                            metadata['files'] = [*request.files.keys()]
                        elif request.is_form_urlencoded:
                            metadata['form'] = request.form.to_dict(False)
                            metadata['files'] = [*request.files.keys()]
                        else:
                            metadata['values'] = request.values.to_dict(False)
                        subsegment.put_metadata('request', metadata, COWORKS_NAMESPACE)
                    except (Exception,):
                        pass

                try:
                    response: CoworksResponse = _view_function(*args, **kwargs)
                except Exception:
                    if subsegment:
                        metadata = {'traceback': traceback.format_exc()}
                        subsegment.put_metadata('exception', metadata, COWORKS_NAMESPACE)
                    raise

                if subsegment:
                    try:
                        metadata = {
                            'status_code': response.status_code,
                            'headers': response.headers,
                        }
                        if response.status_code >= 300:
                            metadata['error'] = response.get_data(as_text=True)
                        subsegment.put_metadata('response', metadata, COWORKS_NAMESPACE)
                    except (Exception,):
                        pass

                return response

            wrapped_fun = update_wrapper(partial(route_captured, view_function), view_function)
            fun_name = getattr(wrapped_fun, '__name__', 'wrapped_fun')
            self._app.view_functions[rule.endpoint] = self._recorder.capture(name=fun_name)(wrapped_fun)

    def capture_exception(self, e):
        self._app.logger.error(f"Event: {request.aws_event}")
        self._app.logger.error(f"Context: {request.aws_context}")
        try:
            subsegment = self._recorder.current_subsegment()
            if subsegment:
                subsegment.add_error_flag()
                subsegment.put_annotation('service', self._app.name)
                subsegment.add_exception(e, traceback.extract_stack())
        except SegmentNotFoundException:
            self._app.logger.debug("Skipped capture exception because the SDK is currently disabled.")
        except Exception:
            raise


def lambda_context_to_json(context):
    return {
        'function_name': context.function_name,
        'function_version': context.function_version,
        'memory_limit_in_mb': context.memory_limit_in_mb,
        'aws_request_id': context.aws_request_id,
        'remaining_time': context.get_remaining_time_in_millis(),
    }


def request_environ_to_json(_request):
    return {
        'in_lambda_context': _request.in_lambda_context,
        'is_multipart': _request.is_multipart,
        'is_form_urlencoded': _request.is_form_urlencoded,
        'max_content_length': _request.max_content_length,
        'endpoint': _request.endpoint,
        'query_string': _request.args,
        'want_form_data_parsed': _request.want_form_data_parsed,
        'script_root': _request.script_root,
        'url_root': _request.url_root,
    }



import sys
import typing as t
from distutils.dir_util import copy_tree
from pathlib import Path

import click
from jinja2 import Environment
from jinja2 import PackageLoader
from jinja2 import select_autoescape

from coworks.version import __version__
from .command import CwsCommand
from .command import no_project_context


@click.command("new", CwsCommand, short_help="Creates a new CoWorks project.")
@click.option('--force', is_flag=True, help="Force creation even if already created.")
@no_project_context
def new_command(force) -> None:
    project_templates = Path(__file__).parent / 'project_templates'

    # Copy project configuration file
    src = project_templates
    project_conf = Path("project.cws.yml")

    if project_conf.exists() and not force:
        click.echo("Project already created. Set 'force' option for recreation.")
        return

    # Copy files
    copy_tree(src.as_posix(), '.')
    Path('template.env').rename('.env')

    # Render project configuration file
    template_loader = PackageLoader(t.cast(str, sys.modules[__name__].__package__))
    jinja_env = Environment(loader=template_loader, autoescape=select_autoescape(['html', 'xml']))
    template = jinja_env.get_template('project.cws.yml')
    output = project_conf
    with output.open("w") as f:
        f.write(template.render({'version': __version__.replace('.', '_')}))

    click.echo('New project created.')



from .data import CursorPagination
from .data import JsonApiDataMixin
from .data import JsonApiDict
from .fetching import FetchingContext
from .fetching import fetching_context
from .jsonapi import JsonApi
from .jsonapi import JsonApiError
from .jsonapi import jsonapi
from .jsonapi import toplevel_from_data
from .jsonapi import toplevel_from_pagination
from .query import ListQuery
from .query import Pagination
from .query import Query

__all__ = [
    'CursorPagination',
    'JsonApiDataMixin',
    'JsonApiDict',
    'FetchingContext',
    'fetching_context',
    'JsonApi',
    'JsonApiError',
    'jsonapi',
    'toplevel_from_data',
    'toplevel_from_pagination',
    'Pagination',
    'Query',
    'ListQuery'
]



import inspect
import os
import re
import types
import typing as t
from functools import update_wrapper
from inspect import Parameter
from inspect import Signature
from inspect import signature
from pathlib import Path
from urllib.parse import parse_qs
from urllib.parse import urlencode
from urllib.parse import urlsplit as urllib_urlsplit
from urllib.parse import urlunsplit as urllib_urlunsplit

import dotenv
from flask import current_app
from flask import json
from flask import make_response
from pydantic import BaseModel
from pydantic import ValidationError
from werkzeug.exceptions import UnprocessableEntity

from .globals import request

if t.TYPE_CHECKING:
    from flask.sansio.scaffold import Scaffold

HTTP_METHODS = ['GET', 'POST', 'PUT', 'DELETE', 'PATCH', 'OPTIONS']

PROJECT_CONFIG_VERSION = 3

DEFAULT_DEV_STAGE = "dev"
DEFAULT_LOCAL_STAGE = "local"
DEFAULT_PROJECT_DIR = "tech"

BIZ_BUCKET_HEADER_KEY: str = 'X-CWS-S3Bucket'
BIZ_KEY_HEADER_KEY: str = 'X-CWS-S3Key'

OPEN_SQUARE_BRACKETED_KWARG_PATTERN = re.compile(r'([a-zA-Z0-9]+)__')
SQUARE_BRACKETED_KWARG_PATTERN = re.compile(r'([a-zA-Z0-9]+)__([a-zA-Z0-9._]+)__')


def create_cws_proxy(scaffold: "Scaffold", func, func_args: list[str], func_kwargs: dict,
                     func_generic_kwargs: str | None):
    """Creates the AWS Lambda proxy function.

    :param scaffold: The Flask or Blueprint object.
    :param func: The initial function proxied.
    :param func_args: The declared function args.
    :param func_kwargs: The declared function kwargs.
    :param func_generic_kwargs: The function generic kwargs if defined (usually **kwargs).
    """

    def proxy(**view_args):
        """
        Adds kwargs parameters to the proxied function.

        :param view_args: Request path parameters.
        """

        def check_keyword_expected(param_name):
            """Alerts when more parameters than expected are defined in request."""
            if func_kwargs and param_name not in func_kwargs:
                _err_msg = f"TypeError: got an unexpected keyword argument '{param_name}'"
                raise UnprocessableEntity(_err_msg)

        def as_fun_params(values: dict, flat=True):
            """Set parameters as simple value or list of values if multiple defined.
           :param values: Dict of values.
           :param flat: If true, the list values of lenth 1 is return as single value.
            """
            params: dict[str, t.Any] = {}
            for k, v in values.items():
                k = remove_brackets(k)

                # if the parameter is a sparse fieldsets
                if splitted := SQUARE_BRACKETED_KWARG_PATTERN.fullmatch(k):
                    for kwarg in func_kwargs:
                        if k.startswith(kwarg) and OPEN_SQUARE_BRACKETED_KWARG_PATTERN.fullmatch(kwarg):
                            k = kwarg
                            if k in params:
                                v = {**params[k], **{splitted.group(2): v}}
                            else:
                                v = {splitted.group(2): v}
                            break

                try:
                    check_keyword_expected(k)
                except UnprocessableEntity:
                    if not func_generic_kwargs:
                        raise
                params[k] = v

            # Flatten single value
            if flat:
                params = {k: v[0] if isinstance(v, list) and len(v) == 1 else v for k, v in params.items()}

            return params

        # Get keyword arguments from request parameters or body
        if func_kwargs or func_generic_kwargs:

            # Adds parameters from query parameters
            if request.method == 'GET':
                get_data = request.values.to_dict(False)
                view_args = dict(**view_args, **as_fun_params(get_data))

            # Adds parameters from body
            elif request.method in ['POST', 'PUT', 'PATCH', 'DELETE']:
                try:
                    if request.is_json:
                        if request.data:
                            post_data = request.json
                            if not isinstance(post_data, dict):
                                if len(func_kwargs) != 1:
                                    msg = f"If request payload is not a dict, there must be only one kwarg {type(post_data)}"
                                    raise UnprocessableEntity(msg)
                                post_data = {next(iter(func_kwargs)): post_data}
                            view_args = {**view_args, **as_fun_params(post_data, False)}
                    elif request.is_multipart:
                        post_data = request.form.to_dict(False)
                        files = request.files.to_dict(False)
                        view_args = {**view_args, **as_fun_params(post_data), **as_fun_params(files)}
                    elif request.is_form_urlencoded:
                        post_data = request.form.to_dict(False)
                        view_args = dict(**view_args, **as_fun_params(post_data))
                    else:
                        post_data = request.values.to_dict(False)
                        view_args = dict(**view_args, **as_fun_params(post_data))
                except Exception as e:
                    raise UnprocessableEntity(str(e))

            else:
                err_msg = f"Keyword arguments are not permitted for {request.method} method."
                raise UnprocessableEntity(err_msg)

        else:
            if not func_args:
                try:
                    if request.content_length:
                        if request.is_json and request.json:
                            err_msg = f"TypeError: got an unexpected arguments (body: {request.json})"
                            raise UnprocessableEntity(err_msg)
                    if request.query_string:
                        err_msg = f"TypeError: got an unexpected arguments (query: {request.query_string!r})"
                        raise UnprocessableEntity(err_msg)
                except Exception as e:
                    current_app.logger.error(f"Should not go here (1) : {str(e)}")
                    current_app.logger.error(f"Should not go here (2) : {request.get_data()}")
                    current_app.logger.error(f"Should not go here (3) : {view_args}")
                    raise

        view_args = as_typed_kwargs(func, view_args)
        result = current_app.ensure_sync(func)(scaffold, **view_args)

        resp = make_response(result) if result is not None else \
            make_response("", 204, {'content-type': 'text/plain'})

        binary_headers = get_cws_annotations(func, "__CWS_BINARY_HEADERS")
        if binary_headers and not request.in_lambda_context:
            resp.headers.update(binary_headers)

        return resp

    return update_wrapper(proxy, func)


def path_join(*args: str) -> str:
    """ Joins given arguments into an entry route.
    Slashes are stripped for each argument.
    """

    reduced = (x.lstrip('/').rstrip('/') for x in args if x)
    return str(Path('/').joinpath(*reduced))[1:]


def make_absolute(route: str, url_prefix: str) -> str:
    """Creates an absolute route.
    """
    path = Path('/')
    if url_prefix:
        path = path / url_prefix
    if route:
        path = path / route
    return str(path)


def trim_underscores(name: str) -> str:
    """Removes starting and ending _ in name.
    """
    if name:
        while name.startswith('_'):
            name = name[1:]
        while name.endswith('_'):
            name = name[:-1]
    return name


def is_arg_parameter(param: Parameter) -> bool:
    """ Checks if the parameter is an arg (not a kwarg)."""
    return param.default == inspect.Parameter.empty


def is_kwarg_parameter(param: Parameter) -> bool:
    """ Checks if the parameter is an arg (not a kwarg)."""
    return param.default != inspect.Parameter.empty


def remove_brackets(name):
    """Removes brackets.
    Parameter like page[number] is passed to the function as page__number__."""
    return name.replace('[', '__').replace(']', '__')


def as_typed_kwargs(func: t.Callable, kwargs: dict):
    def get_typed_value(name: str, parameter_type, val):
        if isinstance(parameter_type, types.UnionType):
            for arg in t.get_args(parameter_type):
                try:
                    return get_typed_value(name, arg, val)
                except (UnprocessableEntity, ValidationError):
                    raise
                except (TypeError, ValueError):
                    pass
            raise TypeError()
        origin = t.get_origin(parameter_type)
        if origin is t.Union:
            for arg in t.get_args(parameter_type):
                return get_typed_value(name, arg, val)
            raise TypeError()
        if origin is list:
            arg = t.get_args(parameter_type)[0]
            if isinstance(val, list):
                return [arg(v) for v in val]
            return [arg(val)]
        if origin is set:
            arg = t.get_args(parameter_type)[0]
            if isinstance(val, list):
                return {arg(v) for v in val}
            return {arg(val)}
        if origin is None:
            if not parameter_type or parameter_type is Signature.empty:
                return val
            if isinstance(val, list):
                msg = f"Multiple values for '{name}' query parameters are not allowed"
                raise UnprocessableEntity(msg)
            if issubclass(parameter_type, bool):
                return str_to_bool(val)
            if issubclass(parameter_type, dict):
                if isinstance(val, str):
                    return json.loads(val)
            if issubclass(parameter_type, BaseModel):
                return parameter_type(**json.loads(val))
            return val if isinstance(val, parameter_type) else parameter_type(val)

    typed_kwargs = {**kwargs}
    try:
        parameters = signature(func).parameters
        for name, value in kwargs.items():
            typed_kwargs[name] = get_typed_value(name, t.cast(Parameter, parameters.get(name)).annotation, value)
    except (UnprocessableEntity, ValidationError):
        raise
    except (TypeError, ValueError) as e:
        raise UnprocessableEntity(str(e))
    except (Exception,):
        pass
    return typed_kwargs


def is_json(mt):
    """Checks if a mime type is json.
    """
    return (
            mt == "application/json"
            or isinstance(mt, str)
            and mt.startswith("application/")
            and mt.endswith("+json")
    )


def str_to_bool(val: str) -> bool:
    return val.lower() in ['true', '1', 'yes']


def get_app_stage():
    """Defined only on deployed microservice or should be set manually."""
    return os.getenv('CWS_STAGE', DEFAULT_DEV_STAGE)


def load_dotenv(stage: str):
    values = {}
    for env_filename in get_env_filenames(stage):
        path = dotenv.find_dotenv(env_filename, usecwd=True)
        if path:
            values.update(dotenv.dotenv_values(path))
    return values


def get_env_filenames(stage):
    return [".env", ".flaskenv", f".env.{stage}", f".flaskenv.{stage}"]


def nr_url(path: str = '', query: dict | None = None, merge_query: bool = False):
    """Combines the arguments into a complete URL as a string.

    :param path: the new path .
    :param query: the query parameters.
    :param merge_query: adds the request query parameters.
    """
    if request.aws_event:
        header = request.aws_event['params']['header']
        if 'forwarded' in header:
            forwarded = {val.split('=')[0]: val.split('=')[1] for val in header['forwarded'].split(';')}
            proto = forwarded['proto']
            host = forwarded['host']
            path = header['x-forwarded-path'] + path
        else:
            proto = header.get('x-forwarded-proto', 'https')
            host = header['host']
    else:
        proto, host, *_ = urllib_urlsplit(request.base_url)

    if merge_query and query:
        if request.aws_event:
            query = {**request.aws_event['params']['querystring'], **query}
        else:
            *_, request_query, _ = urllib_urlsplit(request.url)
            query = {**parse_qs(request_query), **query}

    qs = urlencode(query, doseq=True) if query else ''
    return urllib_urlunsplit([proto, host, path, qs, ''])


def get_cws_annotations(func, key, default=None):
    """Entry function is at least annotated by __CWS_METHOD."""
    if getattr(func, '__CWS_METHOD', None):
        return getattr(func, key, default)
    if getattr(func, '__wrapper__', None):
        return get_cws_annotations(func.__wrapper__, key, default)
    return None



import contextlib
import os
import socket
from unittest.mock import MagicMock

import pytest

fixture_example_dir = os.getenv('EXAMPLE_DIR', 'tests/cws/src')
fixture_samples_docs_dir = os.getenv('SAMPLES_DOCS_DIR', 'samples/docs')


@pytest.fixture
def example_dir(monkeypatch):
    monkeypatch.syspath_prepend(fixture_example_dir)
    monkeypatch.chdir(fixture_example_dir)
    yield fixture_example_dir


@pytest.fixture
def samples_docs_dir(monkeypatch):
    monkeypatch.syspath_prepend(fixture_example_dir)
    monkeypatch.chdir(fixture_samples_docs_dir)
    yield fixture_samples_docs_dir


@pytest.fixture
def progressbar():
    yield MagicMock()


@pytest.fixture
def unused_tcp_port():
    with contextlib.closing(socket.socket()) as sock:
        sock.bind(('localhost', 0))
        return sock.getsockname()[1]


@pytest.fixture
def auth_headers():
    yield {
        'authorization': 'pytest',
    }


@pytest.fixture
def empty_aws_context():
    return object()


def pytest_sessionstart():
    if not os.path.exists(fixture_samples_docs_dir):
        msg = "Undefined samples folder: (environment variable 'SAMPLES_DOCS_DIR' must be redefined)."
        raise pytest.UsageError(msg)
    if not os.path.exists(fixture_example_dir):
        msg = "Undefined example folder: (environment variable 'EXAMPLE_DIR' must be redefined)."
        raise pytest.UsageError(msg)



import os
from collections import defaultdict

from coworks import TechMicroService
from coworks import entry


class EnvTechMS(TechMicroService):
    values = defaultdict(int)
    init_value = None

    def token_authorizer(self, token):
        return True

    @entry
    def get(self, usage="test"):
        """Entrypoint for testing named parameter."""
        return f"Simple microservice for {usage}.\n"

    @entry
    def get_value(self, index):
        """Entrypoint for testing positional parameter."""
        return f"{self.values[index]}\n"

    @entry
    def put_value(self, index, value=0):
        self.values[index] = value
        return value

    @entry
    def get_init(self):
        return f"Initial value is {self.init_value}.\n"

    @entry
    def get_env(self):
        return f"Value of environment variable test is : {os.getenv('TEST')}."



import typing as t
from asyncio import iscoroutine
from functools import update_wrapper
from inspect import Parameter
from inspect import signature

from coworks import TechMicroService
from coworks import request
from flask import current_app
from flask import make_response
from jsonapi_pydantic.v1_0 import Error
from jsonapi_pydantic.v1_0 import ErrorLinks
from jsonapi_pydantic.v1_0 import Link
from jsonapi_pydantic.v1_0 import Relationship
from jsonapi_pydantic.v1_0 import Resource
from jsonapi_pydantic.v1_0 import ResourceIdentifier
from jsonapi_pydantic.v1_0 import TopLevel
from pydantic import ValidationError
from pydantic.networks import HttpUrl
from werkzeug.exceptions import BadRequest
from werkzeug.exceptions import HTTPException
from werkzeug.exceptions import InternalServerError
from werkzeug.exceptions import NotFound

from .data import JsonApiDataMixin
from .data import JsonApiRelationship
from .fetching import create_fetching_context_proxy
from .fetching import fetching_context
from .query import Pagination
from .query import Query


class JsonApiError(Exception):
    """Exception wich will create a JSON:API error."""

    def __init__(self, id_or_error: t.Union[str, Exception, "JsonApiError", list["JsonApiError"]],
                 title: str | None = None, detail: str | None = None, code=None, status=None):
        if isinstance(id_or_error, str):
            code = str(code) or None
            status = str(status) or InternalServerError.code
            try:
                self.errors = [Error(id=id_or_error, code=code, title=title, detail=detail, status=status)]
            except ValidationError as e:
                self.errors = [Error(id=InternalServerError.code, title=str(title), detail=str(e),
                                     status=str(InternalServerError.code))]
        else:
            if isinstance(id_or_error, JsonApiError):
                id_or_error = [id_or_error]
            elif isinstance(id_or_error, Exception):
                self.errors = [Error(id=InternalServerError.code, title=type(id_or_error).__name__,
                                     detail=str(id_or_error), status=str(InternalServerError.code))]
            else:
                self.errors = id_or_error


class JsonApi:
    """Flask's extension implementing JSON:API specification.
    This extension uses the external API of ODOO.

    .. versionchanged:: 0.7.3
        ``env_var_prefix`` parameter may be a dict of bind values.
        GraphQL removed.
    """

    def __init__(self, app=None):
        """
        :param app: Flask application.
        """
        self.app = None

        if app:
            self.init_app(app)

    def init_app(self, app: TechMicroService):
        self.app = app
        handle_user_exception = app.handle_user_exception

        def _handle_user_exception(e):
            if 'application/vnd.api+json' not in request.headers.getlist('accept'):
                return handle_user_exception(e)

            if isinstance(e, ValidationError):
                app.full_logger_error(e)
                errors = [Error(id="", status=BadRequest.code, code=err['type'],
                                links=ErrorLinks(about=err['url']),  # type: ignore[typeddict-item]
                                title=err['msg'], detail=str(err['loc'])) for err in e.errors()]
                errors.append(Error(id="", status=BadRequest.code, title=e.title, detail=str(e)))
                return self._toplevel_error_response(errors, status_code=BadRequest.code)

            try:
                rv = handle_user_exception(e)
                if isinstance(rv, HTTPException):
                    errors = [Error(id='0', title=rv.name, detail=rv.description, status=rv.code)]
                    return self._toplevel_error_response(errors, status_code=rv.code)
            except (Exception,):
                app.full_logger_error(e)
                if isinstance(e, JsonApiError):
                    return self._toplevel_error_response(e.errors)

            errors = [Error(id='0', title=e.__class__.__name__, detail=str(e), status=InternalServerError.code)]
            return self._toplevel_error_response(errors, status_code=InternalServerError.code)

        app.handle_user_exception = _handle_user_exception  # type: ignore[method-assign]

        app.after_request(self._change_content_type)

    def _change_content_type(self, response):
        if 'application/vnd.api+json' not in request.headers.getlist('accept'):
            return response

        response.content_type = 'application/vnd.api+json'
        return response

    def _toplevel_error_response(self, errors, *, status_code=None):
        toplevel = TopLevel(errors=errors).model_dump_json()
        if status_code is None:
            status_code = max((err.status for err in errors))
        response = make_response(toplevel, status_code)
        response.content_type = 'application/vnd.api+json'
        return response


def jsonapi(func):
    """JSON:API decorator.
    Transforms an entry into an SQL entry with result as JSON:API.

    Must have Flask-SQLAlchemy extension installed.
    """

    async def _jsonapi(*args, ensure_one: bool = False, include: str | None = None,
                       fields__: dict | None = None, filters__: dict | None = None, sort: str | None = None,
                       page__number__: int | None = None, page__size__: int | None = None,
                       page__max__: int | None = None,
                       __neorezo__: dict | None = None, **kwargs):
        """

        :param args: entry args.,
                 page: int | None = None, per_page: int | None = None
        :param ensure_one: retrieves only one resource if true (default false).
        :param include:
        :param fields:
        :param filters:
        :param sort:
        :param page:
        :param per_page:
        :param max_per_page:
        :param kwargs: entry kwargs.
        """
        create_fetching_context_proxy(include, fields__, filters__, sort, page__number__, page__size__, page__max__)
        res = func(*args, **kwargs)
        if iscoroutine(res):
            res = await res
        try:
            if isinstance(res, Query):
                _toplevel = get_toplevel_from_query(res, ensure_one)
            elif isinstance(res, TopLevel):
                _toplevel = res
            else:
                errors = [Error(id='0', title="Not a toplevel structure", detail=str(res), status=500)]
                _toplevel = TopLevel(errors=errors)
        except NotFound:
            if not ensure_one:
                _toplevel = TopLevel(data=[])
            else:
                raise NotFound("The requested resource was not found")
        except HTTPException as e:
            errors = [Error(id='0', title=e.name, detail=e.description, status=e.code)]
            _toplevel = TopLevel(errors=errors)
        except Exception as e:
            errors = [Error(id='0', title="Internal server error", detail=str(e), status=500)]
            _toplevel = TopLevel(errors=errors)
        return _toplevel.model_dump_json(exclude_none=True)

    # Adds JSON:API query parameters
    sig = signature(_jsonapi)
    # Removes self and kwargs from jsonapi wrapper
    jsonapi_sig = tuple(sig.parameters.values())[1:-1]
    # Splits variadic keyword to add it at end of the signature
    func_sig1 = tuple(p for p in signature(func).parameters.values() if p.kind != Parameter.VAR_KEYWORD)
    func_sig2 = tuple(p for p in signature(func).parameters.values() if p.kind == Parameter.VAR_KEYWORD)
    sig = sig.replace(parameters=func_sig1 + jsonapi_sig + func_sig2)
    update_wrapper(_jsonapi, func)
    setattr(_jsonapi, '__signature__', sig)
    return _jsonapi


def to_ressource_data(jsonapi_data: JsonApiDataMixin, *,
                      included_prefix: str | None = None, with_relationships: list[str] | None = None) -> dict[
    str, t.Any]:
    """Transform a simple structure data into a jsonapi ressource data.

    Beware : included is a dict of type/id key and jsonapi ressource value
    :param jsonapi_data: the data to transform.
    :param included_prefix: the prefix of the included resources
    """

    # set resource data from basemodel
    _type = jsonapi_data.jsonapi_type
    attrs, rels = jsonapi_data.jsonapi_attributes(fetching_context, with_relationships)
    if 'type' in attrs:
        _type = attrs.pop('type')
    else:
        _type = jsonapi_data.jsonapi_type
    if 'id' in attrs:
        _id = str(attrs.pop('id'))
    else:
        _id = jsonapi_data.jsonapi_id

    # get related resources relationships
    relationships: dict[str, Relationship] = {}
    included: dict[str, dict] = {}
    if included_prefix:
        to_be_included = [k[len(included_prefix):] for k in fetching_context.include or [] if included_prefix in k]
    else:
        to_be_included = [k for k in fetching_context.include or [] if '.' not in k]

    # add relationship and included resources if needed
    for key, rel in rels.items():
        if rel is None:
            continue

        if isinstance(rel, list):
            res_ids = []
            for val in rel:
                res_id = get_resource_identifier(val)
                res_ids.append(res_id)
                add_to_included(included, key, val, to_be_included=to_be_included, included_prefix=included_prefix)
            relationships[key] = Relationship(data=res_ids)
        else:
            res_id = get_resource_identifier(rel)
            add_to_included(included, key, rel, to_be_included=to_be_included, included_prefix=included_prefix)
            relationships[key] = Relationship(data=res_id)

    resource_data = {
        "type": _type,
        "id": _id,
        "lid": None,
        "attributes": attrs,
        "links": get_resource_links(jsonapi_data)
    }

    if relationships:
        resource_data["relationships"] = relationships
    if included:
        resource_data["included"] = included

    return resource_data


def get_toplevel_from_query(query: Query, ensure_one: bool) -> TopLevel:
    """Returns the Toplevel structure from the query."""

    def get_toplevel():
        current_app.logger.debug(str(query))
        if ensure_one:
            all_resources: list[JsonApiDataMixin] = query.all()
            if len(all_resources) != 1:
                raise NotFound("None or more than one resource found and ensure_one parameters was set")
            toplevel = toplevel_from_data(all_resources[0])
        else:
            pagination: Pagination = query.paginate(page=fetching_context.page, per_page=fetching_context.per_page,
                                                    max_per_page=fetching_context.max_per_page)
            toplevel = toplevel_from_pagination(pagination)
        return toplevel

    # connection manager may be iterabl (must be performed asynchronously)
    if isinstance(fetching_context.connection_manager, t.Iterable):
        _toplevels = []
        for connection_manager in fetching_context.connection_manager:
            with connection_manager:
                _toplevels.append(get_toplevel())
        data = [r for tp in _toplevels for r in tp.data]
        included = {i.type + i.id: i for tp in _toplevels if tp.included for i in tp.included}
        if len(_toplevels) == 1:
            meta = _toplevels[0].meta
            links = _toplevels[0].links
        else:
            meta = {"count": len(data)}
            links = {}
        return TopLevel(data=data, included=included.values() if included else None, meta=meta, links=links)

    with fetching_context.connection_manager:
        return get_toplevel()


def toplevel_from_data(res: JsonApiDataMixin):
    """Transform a simple structure data into a toplevel jsonapi.

    :param res: the data to transform.
    """
    data = to_ressource_data(res)
    included: dict = data.pop('included', {})
    resources = Resource(**data)
    included_resources = [Resource(**i) for i in included.values()]
    return TopLevel(data=resources, included=included_resources if included else None)


def toplevel_from_pagination(pagination: Pagination):
    """Transform an iterable pagination into a toplevel jsonapi.

    :param pagination: the data to transform.
    """
    data = [to_ressource_data(d) for d in t.cast(t.Iterable, pagination)]
    included = [d.pop('included') for d in data if 'included' in d]
    resources = [Resource(**d) for d in data]
    included_resources = [Resource(**d) for i in included for d in [*i.values()]]
    toplevel = TopLevel(data=resources, included=included_resources if included else None)
    fetching_context.add_pagination(toplevel, pagination)
    return toplevel


def get_resource_identifier(rel: JsonApiRelationship):
    """ Adds a relationship in the list of relationships from the related model.
    The relationship may not be complete for circular reference and will be completed after in construction.
    """
    if not isinstance(rel, JsonApiRelationship):
        msg = f"Relationship value must be of type JsonApiRelationship, not {rel.__class__}"
        raise InternalServerError(msg)

    type_ = rel.jsonapi_type
    id_ = rel.jsonapi_id
    return ResourceIdentifier(type=type_, id=id_)


def get_resource_links(jsonapi_basemodel) -> dict:
    """Get the links associated to a ressource (from jsonapi_self_link property).
    
    If jsonapi_self_link is a string then there is only one self link.
    """
    self_link = jsonapi_basemodel.jsonapi_self_link
    if isinstance(self_link, str):
        return {'self': Link(href=HttpUrl(jsonapi_basemodel.jsonapi_self_link))}
    if isinstance(self_link, dict):
        return self_link
    raise InternalServerError("Unexpected jsonapi_self_link value")


def add_to_included(included, key, res: JsonApiRelationship, *, to_be_included, included_prefix):
    """Adds the resource defined at key to the included list of resource.
    
    :param included: list of included resources to increment (if not already inside).
    :param key: the key where the resource is in the parent resource.
    :param res: the relationship to include.
    :param to_be_included: the list of keys to include in the included list of resources.
    :param included_prefix: dot separated path in resource.
    """
    res_key = res.jsonapi_type + res.jsonapi_id
    if key in to_be_included and res_key not in included:
        new_included_prefix = f"{included_prefix}{key}." if included_prefix else f"{key}."
        with_relationships = get_relationships_to_add_in_included(new_included_prefix)
        if res.resource_value:
            res_included = to_ressource_data(res.resource_value, included_prefix=new_included_prefix,
                                             with_relationships=with_relationships)
            included[res_key] = res_included

            # Moves included ressources defined in this included ressource
            if 'included' in res_included:
                for k, v in res_included.pop('included').items():
                    included[k] = v


def get_relationships_to_add_in_included(new_included_prefix):
    """Returns the relationships to add from the include fetching context.
    Get only attributesd key with the new included prefix and no more.

    :param new_included_prefix: the new prefix of ressources."""
    prefixed_include = [i[len(new_included_prefix):] for i in fetching_context.include
                        if i.startswith(new_included_prefix)]
    return [i for i in prefixed_include if '.' not in i]



import os
from unittest import mock

from flask import json

from coworks import TechMicroService
from coworks import entry
from ..event import get_event


class ContentMS(TechMicroService):

    @entry
    def get(self):
        return "test"

    @entry
    def get_json(self):
        return {'text': 'value', 'int': 1}

    @entry
    def post(self, text=None, context=None, files=None):
        if files:
            if type(files) is not list:
                files = [files]
            return f"post {text}, {context} and {[f.file.name for f in files]}"
        return f"post {text}, {context}"

    @entry(binary_headers={'content-type': 'application/octet'}, no_auth=True)
    def get_binary(self):
        return b"test"

    @entry(binary_headers={'content-type': 'application/pdf'}, no_auth=True)
    def get_content_type(self):
        return b"test"

    @entry(binary_headers={'content-type': 'application/octet'}, no_auth=True)
    def get_no_auth(self):
        return b"test"


@mock.patch.dict(os.environ, {"TOKEN": "token"})
class TestClass:
    def test_default_content_type(self):
        app = ContentMS()
        with app.test_client() as c:
            headers = {'Authorization': 'token'}
            response = c.get('/', headers=headers)
            assert response.status_code == 200
            assert response.is_json
            assert response.headers['Content-Type'] == 'application/json'
            assert response.get_data(as_text=True) == 'test'

    def test_json_content_type(self):
        app = ContentMS()
        with app.test_client() as c:
            headers = {'Accept': 'application/json', 'Authorization': 'token'}
            response = c.get('/', headers=headers)
            assert response.status_code == 200
            assert response.is_json
            assert response.headers['Content-Type'] == 'application/json'
            assert response.get_data(as_text=True) == 'test'

    def test_text_content_type(self):
        app = ContentMS()
        with app.test_client() as c:
            headers = {'Accept': 'text/plain', 'Authorization': 'token'}
            response = c.get('/', headers=headers)
            assert response.status_code == 200
            assert response.is_json
            assert response.headers['Content-Type'] == 'application/json'
            assert response.get_data(as_text=True) == 'test'

    def test_text_api(self):
        app = ContentMS()
        with app.test_client() as c:
            headers = {'Authorization': 'token'}
            response = c.get('/json', headers=headers)
            assert response.status_code == 200
            assert response.is_json
            assert response.headers['Content-Type'] == 'application/json'
            assert response.json == {"int": 1, "text": "value"}

            headers = {'Accept': 'application/json', 'Authorization': 'token'}
            response = c.get('/json', headers=headers)
            assert response.status_code == 200
            assert response.is_json
            assert response.headers['Content-Type'] == 'application/json'
            assert response.json == {"int": 1, "text": "value"}

            headers = {'Accept': 'text/plain', 'Authorization': 'token'}
            response = c.get('/json', headers=headers)
            assert response.status_code == 200
            assert response.is_json
            assert response.headers['Content-Type'] == 'application/json'
            assert json.loads(response.get_data(as_text=True)) == {"text": "value", "int": 1}

    def test_binary_content_type(self, empty_aws_context):
        app = ContentMS()
        with app.test_client() as c:
            headers = {'Accept': 'img/webp', 'Authorization': 'token'}
            response = app(get_event('/binary', 'get', headers=headers), empty_aws_context)
            # assert type(response) == str
            # assert base64.b64decode(str(response)) == b"test"

    def test_content_type(self, empty_aws_context):
        app = ContentMS()
        with app.test_client() as c:
            headers = {'Accept': 'img/webp', 'Authorization': 'token'}
            response = app(get_event('/content/type', 'get', headers=headers), empty_aws_context)
            # assert type(response) == str
            # assert base64.b64decode(str(response)) == b"test"



import os

import boto3


class Boto3Mixin:

    def __init__(self, service, profile_name=None, env_var_access_key='aws_access_key_id',
                 env_var_secret_key='aws_secret_access_key', env_var_region='aws_region',
                 **kwargs):
        self.__session__ = self.__client__ = None
        self.__service = service
        self.__env_var_access_key = env_var_access_key
        self.__env_var_secret_key = env_var_secret_key
        self.__env_var_region = env_var_region

        self.__profile_name = profile_name

    @property
    def aws_access_key(self):
        value = os.getenv(self.__env_var_access_key)
        if not value:
            raise RuntimeError(f"{self.__env_var_access_key} not defined in environment")
        return value

    @property
    def aws_secret_access_key(self):
        value = os.getenv(self.__env_var_secret_key)
        if not value:
            raise RuntimeError(f"{self.__env_var_secret_key} not defined in environment")
        return value

    @property
    def region_name(self):
        if self.__session__:
            return self.__session__.region_name

        value = os.getenv(self.__env_var_region)
        if not value and self.aws_access_key and self.aws_secret_access_key:
            raise RuntimeError(f"{self.__env_var_region} not defined in environment")
        return value

    @property
    def client(self):
        if self.__client__ is None:
            self.__client__ = self.__session.client(self.__service)
        return self.__client__

    @property
    def __session(self):
        if self.__session__ is None:
            if self.__profile_name is not None:
                try:
                    self.__session__ = boto3.Session(profile_name=self.__profile_name)
                except Exception:
                    raise RuntimeError(f"Cannot create session for profile {self.__profile_name}.")
            else:
                access_key = self.aws_access_key
                secret_key = self.aws_secret_access_key
                region_name = self.region_name
                try:
                    self.__session__ = boto3.Session(access_key, secret_key, region_name=region_name)
                except Exception:
                    raise RuntimeError(
                        f"Cannot create session for key {access_key}, secret {secret_key} in {region_name}."
                    )
        return self.__session__


class AwsS3Session(Boto3Mixin):

    def __init__(self, **kwargs):
        super().__init__('s3', **kwargs)



from flask import json


def get_event(entry_path, method, entry_path_parameters=None, params=None, body=None, headers=None):
    headers = headers or {
        'accept': '*/*',
        'authorization': 'token',
        'content-type': 'application/json'
    }
    return {
        'type': 'LAMBDA',
        'resource': entry_path,
        'path': entry_path,
        'httpMethod': method.upper(),
        'headers': {
            'accept-encoding': 'gzip, deflate, br',
            'cache-control': 'max-age=0',
            'cookie': 'session=70b58773-af3e-4153-a5ab-5356481ea87e',
            'host': 'zb53tzphr8.execute-api.eu-west-1.amazonaws.com',
            'origin': 'https://neorezo.io',
            'sec-ch-ua': '" Not;A Brand";v="99", "Google Chrome";v="97", "Chromium";v="97"',
            'sec-ch-ua-mobile': '?0',
            'sec-ch-ua-platform': '"Linux"',
            'sec-fetch-dest': 'document',
            'sec-fetch-mode': 'navigate',
            'sec-fetch-site': 'same-origin',
            'sec-fetch-user': '?1',
            'upgrade-insecure-requests': '1',
            'user-agent': 'Amazon CloudFront',
            'via': '2.0 4e4ca876a59e9f2e22ec751bbab5f282.cloudfront.net (CloudFront)',
            'x-amz-cf-id': '0gvkeT2_kuutjZHh075zWXA0pfqm-afERp4UltS3XBAguLOI4tdckw==',
            'x-amzn-trace-id': 'Root=1-620f51ee-3f8b28746271cbea29b0084e',
            'x-forwarded-for': '91.170.42.203, 130.176.185.118',
            'x-forwarded-port': '443',
            'x-forwarded-proto': 'https',
            **headers
        },
        'multiValueHeaders': {},
        'body': json.dumps(body or {}),
        'queryStringParameters': {},
        'multiValueQueryStringParameters': params or {},
        'pathParameters': {},
        'stageVariables': None,
        'isBase64Encoded': False,
        'entryPathParameters': entry_path_parameters or {},
        'requestContext': {
            'httpMethod': method.upper(),
            'resourceId': 'fp5ol74tr7',
            'entryPath': entry_path,
            'extendedRequestId': 'EktgyFweDoEFabw=',
            'requestTime': '24/Aug/2021:13:37:08 +0000',
            'path': f"/dev{entry_path}",
            'accountId': '935392763270',
            'protocol': 'HTTP/1.1',
            'stage': 'dev',
            'domainPrefix': 'htzd2rneg1',
            'requestTimeEpoch': 1629812228818,
            'requestId': '2fa7f00a-58fe-4f46-a829-1fee00898e42',
            'domainName': 'htzd2rneg1.execute-api.eu-west-1.amazonaws.com',
            'apiId': 'htzd2rneg1'
        },
        'params': {
            'path': {},
            'querystring': {},
            'header': {'Accept': '*/*', 'Authorization': 'token', 'CloudFront-Forwarded-Proto': 'https',
                       'CloudFront-Is-Desktop-Viewer': 'true', 'CloudFront-Is-Mobile-Viewer': 'false',
                       'CloudFront-Is-SmartTV-Viewer': 'false', 'CloudFront-Is-Tablet-Viewer': 'false',
                       'CloudFront-Viewer-Country': 'FR', 'content-type': 'application/json',
                       'Cookie': 'session=70b58773-af3e-4153-a5ab-5356481ea87e',
                       'Host': 'htzd2rneg1.execute-api.eu-west-1.amazonaws.com', 'User-Agent': 'insomnia/2021.4.1',
                       'Via': '2.0 4dd111c814b0b5cf8bf82e59008da625.cloudfront.net (CloudFront)',
                       'X-Amz-Cf-Id': 'ka1hbQCSUOZ-d0VQYuE_gtF4icy443t7kP3UGsDLZDF_5QyTX13FoQ==',
                       'X-Amzn-Trace-Id': 'Root=1-6124f604-3fb9457c7489ebf14ed0f8f6',
                       'X-Forwarded-For': '78.234.174.193, 130.176.152.165', 'X-Forwarded-Port': '443',
                       'X-Forwarded-Proto': 'https'}},
        'context': {
            'resourceId': 'fp5ol74tr7',
            'authorizer': '',
            'resourcePath': '/',
            'httpMethod': 'GET',
            'extendedRequestId': 'EktgyFweDoEFabw=',
            'requestTime': '24/Aug/2021:13:37:08 +0000',
            'path': '/dev/',
            'accountId': '935392763270',
            'protocol': 'HTTP/1.1',
            'requestOverride': '',
            'stage': 'dev',
            'domainPrefix': 'htzd2rneg1',
            'requestTimeEpoch': '1629812228818',
            'requestId': '2fa7f00a-58fe-4f46-a829-1fee00898e42',
            'identity': '',
            'domainName': 'htzd2rneg1.execute-api.eu-west-1.amazonaws.com',
            'responseOverride': '',
            'apiId': 'htzd2rneg1'
        }
    }



import typing as t

from coworks import TechMicroService
from coworks import entry


class TypedMS(TechMicroService):

    def _check_token(self):
        """No check."""

    @entry
    def get(self, i: int):
        return ("ok", 200) if type(i) is int else ("not ok", 400)

    @entry
    def get_(self, i: int = 0):
        return ("ok", 200) if type(i) is int else ("not ok", 400)

    @entry
    def get_bool(self, i: bool):
        if type(i) is bool:
            return ("true", 200) if i else ("false", 200)
        return "not ok", 400

    @entry
    def get_bool_(self, i: bool = None):
        if type(i) is bool:
            return ("true", 200) if i else ("false", 200)
        return "not ok", 400

    @entry
    def get_union(self, i: t.Union[int, int] = 0):
        return ("ok", 200) if type(i) is int else ("not ok", 400)

    @entry
    def get_list(self, i: list[int] = 0):
        return ("ok", 200) if type(i) is list else ("not ok", 400)

    @entry
    def post(self, i: t.Union[int, int] = 0):
        return ("ok", 200) if type(i) is int else ("not ok", 400)


class TestClass:

    def test_base_type(self):
        app = TypedMS()

        with app.test_client() as c:
            response = c.get('/1')
            assert response.status_code == 200

            response = c.get('/?i=1')
            assert response.status_code == 200

            response = c.post('/', json={'i': 1})
            assert response.status_code == 200

            response = c.post('/', json={'i': '1'})
            assert response.status_code == 200

            response = c.post('/', json={'i': 'abc'})
            assert response.status_code == 422

    def test_bool(self):
        app = TypedMS()

        with app.test_client() as c:
            response = c.get('/bool/true')
            assert response.status_code == 200
            assert response.get_data(as_text=True) == "true"

            response = c.get('/bool/1')
            assert response.status_code == 200
            assert response.get_data(as_text=True) == "true"

            response = c.get('/bool/false')
            assert response.status_code == 200
            assert response.get_data(as_text=True) == "false"

            response = c.get('/bool/0')
            assert response.status_code == 200
            assert response.get_data(as_text=True) == "false"

            response = c.get('/bool?i=true')
            assert response.status_code == 200
            assert response.get_data(as_text=True) == "true"

            response = c.get('/bool?i=1')
            assert response.status_code == 200
            assert response.get_data(as_text=True) == "true"

            response = c.get('/bool?i=false')
            assert response.status_code == 200
            assert response.get_data(as_text=True) == "false"

            response = c.get('/bool?i=0')
            assert response.status_code == 200
            assert response.get_data(as_text=True) == "false"


    def test_union_type(self):
        app = TypedMS()

        with app.test_client() as c:
            response = c.get('/union?i=1')
            assert response.status_code == 200

            response = c.get('/union?i=1&i=2')
            assert response.status_code == 422

            response = c.get('/union?i=abc')
            assert response.status_code == 422

    def test_list_type(self):
        app = TypedMS()

        with app.test_client() as c:
            response = c.get('/list?i=1')
            assert response.status_code == 200

            response = c.get('/list?i=1&i=2')
            assert response.status_code == 200

            response = c.get('/list?i=abc')
            assert response.status_code == 422



import base64
import typing as t
from json import JSONDecodeError
from json import loads

import requests
from airflow.exceptions import AirflowFailException
from airflow.models import Variable
from airflow.models.baseoperator import BaseOperator
from airflow.operators.branch import BaseBranchOperator
from airflow.providers.amazon.aws.hooks.s3 import S3Hook
from airflow.providers.http.hooks.http import HttpHook

XCOM_DEFAULT_KEY = 'return_value'
XCOM_CWS_BUCKET = 'bucket'  # the AWS S3 bucket where the coworks techmicroservice result is stored
XCOM_CWS_KEY = 'key'  # the AWS S3 key where the coworks techmicroservice result is stored
XCOM_CWS_NAME = 'cws_name'
XCOM_STATUS_CODE = 'status_code'


class TechMicroServiceOperator(BaseOperator):
    template_fields = ["cws_name", "headers", "entry", "query_params", "json", "data"]

    def __init__(self, *, cws_name: str = None, entry: str = '/', method: str = 'get', no_auth: bool = False,
                 query_params: t.Union[dict, str] = None, json: t.Union[dict, str] = None,
                 data: t.Union[dict, str] = None,
                 stage: str = None, api_id: str = None, token: str = None,
                 raise_errors: bool = True, raise_400_errors: bool = True,
                 accept: str = 'application/json', headers: dict = None, log_response: bool = False,
                 directory_conn_id: str = 'coworks_directory', asynchronous: bool = False,
                 xcom_push: bool = True, json_result: bool = True,
                 multiple_outputs_transformer: t.Callable[[dict], dict] = None,
                 **kwargs) -> None:
        """Microservice operator.
        The tech microservice may be called from its name or from it api_id, stage and token.

        :param cws_name: the tech microservice name.
        :param entry: the route entry.
        :param method: the route method ('GET', 'POST').
        :param no_auth: set to 'True' if no authorization is needded (default 'False').
        :param query_params: query parameters for GET method.
        :param json: dict data for POST method.
        :param data: object to send in the body for POST method
        :param stage: the microservice stage (default 'dev' if 'cws_name' not defined).
        :param api_id: APIGateway id (must be defined if no 'cws_name').
        :param token: Authorization token (must be defined if auth and no 'cws_name').
        :param directory_conn_id: connection defined for the directory service (default 'coworks_directory').
        :param asynchronous: asynchronous call (default False).
        :param xcom_push: pushes result in XCom (default True).
        :param json_result: returns a JSON value in 'return_value' (default True).
        :param raise_errors: raise error on client errors (default True).
        :param raise_400_errors: raise error on client 400 errors (default True).
        :param accept: accept header value (default 'application/json').
        :param headers: specific header values added to default ones and may override them (default {}).
        :param log_response: trace result content (default False).
        :param multiple_outputs_transformer: if defined, return a multi-output XCOM after tranformation.

         .. versionchanged:: 0.8.4
            Added the ``multiple_outputs_transformer`` parameter.
        """
        super().__init__(**kwargs)
        self.cws_name = cws_name
        self.entry = entry.lstrip('/')
        self.method = method.lower()
        self.no_auth = no_auth
        self.log_response = log_response
        self.query_params = query_params
        self.json = json
        self.data = data
        self.stage = stage
        self.api_id = api_id
        self.token = token
        self.directory_conn_id = directory_conn_id
        self.asynchronous = asynchronous
        self.xcom_push_flag = xcom_push
        self.json_result = json_result
        self.raise_errors = raise_errors
        self.raise_400_errors = raise_400_errors
        self.multiple_outputs_transformer = multiple_outputs_transformer
        self.accept = accept
        self.headers = headers
        self._url = self._bucket = self._key = self.__headers = None

    def pre_execute(self, context):
        """Gets url and token from name or parameters.

        Done only before execution not on DAG loading.
        """
        dag_run = context['dag_run']
        start_date = dag_run.start_date.timestamp()
        trace_id = f"Root=1-{hex(int(start_date))[2:]}-{f'cws{dag_run.id:0>9}'.encode().hex()}"
        self._headers['x-amzn-trace-id'] = trace_id
        self.log.info(f"Cws trace: {trace_id}")

        if self.cws_name:
            http = HttpHook('get', http_conn_id=self.directory_conn_id)
            self.log.info("Calling CoWorks directory")
            data = {'stage': self.stage} if self.stage else {}
            response = http.run(self.cws_name, data=data)
            coworks_data = loads(response.text)
            self.token = coworks_data['token']
            self._url = f"{coworks_data['url']}/{self.entry}"
        else:
            self._url = self.url

        if self.asynchronous:
            ti = context['ti']
            self._bucket = 'coworks-airflow'
            self._key = f"s3/{dag_run.dag_id}/{ti.task_id}/{ti.job_id}"

            self._headers['InvocationType'] = 'Event'
            self._headers['X-CWS-S3Bucket'] = self._bucket
            self._headers['X-CWS-S3Key'] = self._key
            self.log.info(f"Result stored in 's3://{self._bucket}/{self._key}'")

        if not self.no_auth:
            self._headers['Authorization'] = self.token

    def execute(self, context):
        """Call TechMicroService.
        """
        resp = self._call_cws(context)

        if self.raise_errors:
            if (self.raise_400_errors and resp.status_code >= 400) or resp.status_code >= 500:
                if self.xcom_push_flag:
                    self._push_response(context, resp)

                msg = f"The TechMicroService {self.cws_name} had an error {resp.status_code}!"
                self.log.error(msg)
                raise AirflowFailException(msg)

        if self.xcom_push_flag:
            self._push_response(context, resp)

    @property
    def url(self):
        """Default url construction."""
        return f'https://{self.api_id}.execute-api.eu-west-1.amazonaws.com/{self.stage}/{self.entry}'

    @property
    def default_headers(self):
        """Default headers values."""
        return {
            'Content-Type': "application/json",
            'Accept': self.accept,
        }

    @property
    def _headers(self):
        if self.__headers is None:
            self.__headers = self.default_headers
            if self.headers:
                self.__headers.update(self.headers)
        return self.__headers

    def _call_cws(self, context):
        """Method used by operator and sensor."""
        self.log.info(f"Calling {self.method.upper()} method to {self._url}")
        resp = requests.request(
            self.method.upper(), self._url, headers=self._headers,
            params=self.query_params, json=self.json, data=self.data
        )
        self.log.info(f"Resulting status code : {resp.status_code}")
        return resp

    def _push_response(self, context, resp):
        # Return values or store file information
        self.xcom_push(context, XCOM_CWS_NAME, self.cws_name or self.api_id)
        if self.asynchronous:
            self.xcom_push(context, XCOM_CWS_BUCKET, self._bucket)
            self.xcom_push(context, XCOM_CWS_KEY, self._key)
        else:
            if self.json_result:
                try:
                    returned_value = resp.json() if resp.content else {}
                except JSONDecodeError:
                    self.log.error(f"Not a JSON value: {resp.text}'")
                    returned_value = resp.text
            else:
                returned_value = resp.text

            if self.log_response:
                self.log.info(returned_value)

            self.xcom_push(context, XCOM_STATUS_CODE, resp.status_code)
            if self.multiple_outputs_transformer:
                for key, value in self.multiple_outputs_transformer(returned_value):
                    self.xcom_push(context, key, value)

            self.xcom_push(context, XCOM_DEFAULT_KEY, returned_value)


class AsyncTechServicePullOperator(BaseOperator):

    def __init__(self, *, cws_task_id: str = None, aws_conn_id: str = 'aws_s3',
                 raise_errors: bool = True, raise_400_errors: bool = True,
                 xcom_push: bool = True, **kwargs) -> None:
        """Pull in XCom a microservice result when its was called asynchronously.

        :param cws_task_id: the tech microservice called asynchronously.
        :param aws_conn_id: aws connection (default 'aws_s3').
        :param raise_errors: raise error on client errors (default True).
        :param raise_400_errors: raise error on client 400 errors (default True).
        :param xcom_push: pushes result in XCom (default True).

        .. versionchanged:: 0.8.4
            Added the ``xcom_push`` parameter.
        """

        super().__init__(**kwargs)
        self.cws_task_id = cws_task_id
        self.aws_conn_id = aws_conn_id
        self.raise_errors = raise_errors
        self.raise_400_errors = raise_400_errors
        self.xcom_push_flag = xcom_push

    def execute(self, context):
        ti = context["ti"]

        # Reads bucket file content
        bucket_name = ti.xcom_pull(task_ids=self.cws_task_id, key=XCOM_CWS_BUCKET)
        key = ti.xcom_pull(task_ids=self.cws_task_id, key=XCOM_CWS_KEY)
        s3 = S3Hook(aws_conn_id=self.aws_conn_id)
        file = s3.download_file(key, bucket_name=bucket_name)
        with open(file, "r") as myfile:
            data = myfile.read()
        payload = loads(data)

        status_code = payload['statusCode']
        self.xcom_push(context, XCOM_STATUS_CODE, status_code)

        if self.raise_errors:
            if (self.raise_400_errors and status_code >= 400) or status_code >= 500:
                if self.xcom_push_flag:
                    self.xcom_push(context, XCOM_DEFAULT_KEY, payload.get('body'))
                self.log.error(f"Error: {payload['body']}'")
                raise AirflowFailException(f"TechMicroService doesn't complete successfully: {status_code}")

        if self.xcom_push_flag:
            if payload['isBase64Encoded']:
                return base64.b64decode(payload['body'])
            return payload['body']


class BranchTechMicroServiceOperator(BaseBranchOperator):
    """Branch operator based on TechMicroservice call.

    :param cws_task_id: the tech microservice task_id result tested to determine the branch.
    :param on_failure: the task_ids in case of status code returned >= 400.
    :param on_no_content: the task_ids in case of status code returned == 204.
    :param response_check: function evaluated in case of success and on_check defined.
    :param on_check: the task_ids in case of status code returned == 200 and response checked.
    :param on_success: the task_ids in case of status code returned == 200 and on check branch not selected.
    """

    def __init__(self, *, cws_task_id: str = None, on_success: t.Union[str, t.Iterable[str]] = None,
                 on_failure: t.Union[str, t.Iterable[str]] = None, on_no_content: t.Union[str, t.Iterable[str]] = None,
                 response_check: t.Optional[t.Callable[..., bool]] = None, on_check: str = None, **kwargs) -> None:
        super().__init__(**kwargs)
        self.cws_task_id = cws_task_id
        self.on_success = on_success
        self.on_failure = on_failure
        self.on_no_content = on_no_content
        self.response_check = response_check
        self.on_check = on_check

    def choose_branch(self, context):
        status_code = int(context['ti'].xcom_pull(task_ids=self.cws_task_id, key=XCOM_STATUS_CODE))
        self.log.info(f"TechMS {self.cws_task_id} returned code : {status_code}")
        if self.on_failure and status_code >= 400:
            return self.on_failure
        if self.on_no_content and status_code == 204:
            return self.on_no_content
        if self.on_check and self.response_check:
            text = int(context['ti'].xcom_pull(task_ids=self.cws_task_id))
            if self.response_check(text):
                return self.on_check
        return self.on_success


class NeoRezoServiceOperator(TechMicroServiceOperator):
    """This operator is defined as an example of how to use TechMicroServiceOperator.

     This model used JSONAPI extension."""

    def __init__(self, *, module: str = None, service: str = None, accept: str = 'application/vnd.api+json', **kwargs):
        super().__init__(accept=accept, **kwargs)
        self.module = module
        self.service = service

    @property
    def url(self):
        path = f"{self.module}/{self.service}/{self.entry}"
        if self.stage == 'dev':
            return f'https://dev.jsonapi.neorezo.io/{path}'
        return f'https://jsonapi.neorezo.io/{path}'

    @property
    def default_headers(self):
        nr_jwt = Variable.get("BIZ_NR_JWT")
        return {
            'Content-Type': "application/json",
            'Accept': self.accept,
            'X-NR-JWT': nr_jwt
        }



import click
from flask.cli import pass_script_info

from tests.cws.src.app import EnvTechMS


@click.command("test", short_help="Test custom command.")
@click.option('-a', required=True)
@click.option('--b')
@click.pass_context
@pass_script_info
def cmd(info, ctx, a, b):
    cws_app = info.load_app()
    assert cws_app is not None
    assert cws_app.config is not None
    if a:
        print(f"test command with a={a}/", end='')
    print(f"test command with b={b}", end='', flush=True)


@click.command("test", short_help="Test custom command.")
@click.option('-a', required=True)
@click.option('--b')
@click.pass_context
@pass_script_info
def cmd1(info, ctx, a, b):
    cws_app = info.load_app()
    assert cws_app is not None
    assert cws_app.config is not None
    if a:
        print(f"test command v1 with a={a}/", end='')
    print(f"test command v1 with b={b}", end='', flush=True)


app = EnvTechMS()



from coworks import Blueprint, entry
from tests.coworks.ms import TechMS


class EmptyBlueprint(Blueprint):

    @entry
    def get(self):
        return f"blueprint test"


class TestClass:

    def test_as_root(self):
        app = TechMS()
        app.register_blueprint(EmptyBlueprint("bp"))
        with app.test_client() as c:
            response = c.get('/', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == 'blueprint test'
            assert app.routes == ['/']

    def test_with_prefix(self):
        app = TechMS()
        app.register_blueprint(EmptyBlueprint("bp"), url_prefix='bp')
        with app.test_client() as c:
            response = c.get('/bp', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == 'blueprint test'
            assert app.routes == ['/bp']



import multiprocessing
import os
import time
from unittest import mock

import requests

from coworks.cws.client import CwsScriptInfo


class TestClass:

    @mock.patch.dict(os.environ, {"FLASK_RUN_FROM_CLI": "false"})
    def test_run_first_no_token(self, samples_docs_dir, unused_tcp_port):
        info = CwsScriptInfo(project_dir='tech')
        info.app_import_path = "first:app"
        app = info.load_app()
        server = multiprocessing.Process(target=run_server, args=(app, unused_tcp_port), daemon=True)
        server.start()
        counter = 1
        time.sleep(counter)
        while not server.is_alive() and counter < 10:
            time.sleep(counter)
            counter += 1
        response = requests.get(f'http://localhost:{unused_tcp_port}/')
        assert response.status_code == 403
        server.terminate()

    @mock.patch.dict(os.environ, {"FLASK_RUN_FROM_CLI": "false"})
    def test_run_first_wrong_token(self, samples_docs_dir, unused_tcp_port):
        info = CwsScriptInfo(project_dir='tech')
        info.app_import_path = "first:app"
        app = info.load_app()
        server = multiprocessing.Process(target=run_server, args=(app, unused_tcp_port), daemon=True)
        server.start()
        counter = 1
        time.sleep(counter)
        while not server.is_alive() and counter < 10:
            time.sleep(counter)
            counter += 1
        response = requests.get(f'http://localhost:{unused_tcp_port}/', headers={'Authorization': 'wrong'})
        assert response.status_code == 403

    @mock.patch.dict(os.environ, {"FLASK_RUN_FROM_CLI": "false"})
    def test_run_first(self, samples_docs_dir, unused_tcp_port):
        info = CwsScriptInfo(project_dir='tech')
        info.app_import_path = "first:app"
        app = info.load_app()
        server = multiprocessing.Process(target=run_server, args=(app, unused_tcp_port), daemon=True)
        server.start()
        counter = 1
        time.sleep(counter)
        while not server.is_alive() and counter < 10:
            time.sleep(counter)
            counter += 1
        response = requests.get(f'http://localhost:{unused_tcp_port}/', headers={'Authorization': "token"})
        assert response.text == "Stored value 0.\n"
        response = requests.post(f'http://localhost:{unused_tcp_port}/', json={'value': 1},
                                 headers={'Authorization': "token"})
        assert response.text == "Value stored (1).\n"
        response = requests.get(f'http://localhost:{unused_tcp_port}/', headers={'Authorization': "token"})
        assert response.text == "Stored value 1.\n"
        server.terminate()


def run_server(app, port):
    print(f"Server starting on port {port}")
    app.run(host='localhost', port=port, use_reloader=False, debug=False)



import io

from werkzeug.middleware.profiler import ProfilerMiddleware

from coworks import Blueprint
from coworks import entry


class Profiler(Blueprint):

    def __init__(self, app, output=None, **kwargs):
        super().__init__(**kwargs)
        self.output = output or io.StringIO()
        app.wsgi_app = ProfilerMiddleware(app.wsgi_app, stream=self.output)

    @entry
    def get(self):
        profile = self.output.getvalue()
        self.output.seek(0)
        return profile



import typing as t

from airflow.models import BaseOperator
from airflow.operators.python import PythonOperator
from airflow.utils.task_group import TaskGroup
from airflow.utils.trigger_rule import TriggerRule

from coworks.biz.operators import AsyncTechServicePullOperator
from coworks.biz.operators import NeoRezoServiceOperator
from coworks.biz.operators import TechMicroServiceOperator
from coworks.biz.sensors import AsyncTechMicroServiceSensor


class CoworksTaskGroup(TaskGroup):
    """ Asynchronous tasks group.

    .. versionchanged:: 0.8.4
        Added the ``start_id``, ``end_id`` properties.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.transformer_task: t.Optional[BaseOperator] = None
        self.call_task: t.Optional[BaseOperator] = None
        self.wait_task: t.Optional[BaseOperator] = None
        self.read_task: t.Optional[BaseOperator] = None

    @property
    def start_id(self):
        return f'{self._group_id}.transformer' if self.transformer_task else f'{self._group_id}.call'

    @property
    def end_id(self):
        return f'{self._group_id}.read' if self.read_task else f'{self._group_id}.wait'

    @property
    def output(self):
        return self.read_task.output


def TechMicroServiceAsyncGroup(group_id: str, transformer: t.Callable = None, read: bool = True,
                               op_args: t.Optional[t.Collection[t.Any]] = None,
                               op_kwargs: t.Optional[t.Mapping[str, t.Any]] = None,
                               method: str = 'get', timeout: int = 900,
                               raise_errors: bool = True, raise_400_errors: bool = True,
                               xcom_push: bool = True, trigger_rule=TriggerRule.ALL_SUCCESS,
                               **tech_kwargs):
    """Task group to allow asynchronous call of a TechMicroService.

    The returned value is defined in the task_id : '{group_id}.read'.

    :param group_id: group id.
    :param transformer: python transformer function.
    :param read: do read task.
    :param op_args: python transformer args.
    :param op_kwargs: python transformer kwargs.
    :param method: microservice method called.
    :param raise_errors: raise error on client errors (default True).
    :param raise_400_errors: raise error on client 400 errors (default True).
    :param timeout: asynchronous call timeout.
    :param trigger_rule: trigger rule to be set on first task.
    :param xcom_push:  pushes result in XCom (default True).

    .. versionchanged:: 0.8.4
        Added the ``read`` parameter.
    .. versionchanged:: 0.8.4
        Added the ``trigger_rule`` parameter.
    .. versionchanged:: 0.8.0
        Added the ``transformer`` parameter.
    """
    with CoworksTaskGroup(group_id=group_id) as tg:
        if transformer:
            tg.transformer_task = PythonOperator(
                task_id='transformer',
                python_callable=transformer,
                op_args=op_args,
                op_kwargs=op_kwargs,
                trigger_rule=trigger_rule
            )

            if method.lower() == 'get':
                if 'query_params' in tech_kwargs:
                    msg = "Calling transformer method with already query_params parameter call"
                    tg.transformer_task.log.warning(msg)
                tech_kwargs['query_params'] = tg.transformer_task.output
            else:
                if 'json' in tech_kwargs:
                    tg.transformer_task.log.warning("Calling transformer method with already json parameter call")
                tech_kwargs['json'] = tg.transformer_task.output

        tg.call_task = TechMicroServiceOperator(
            task_id="call",
            asynchronous=True,
            method=method,
            raise_errors=raise_errors,
            raise_400_errors=raise_400_errors,
            trigger_rule=trigger_rule if not transformer else TriggerRule.ALL_SUCCESS,
            **tech_kwargs,
        )

        tg.wait_task = AsyncTechMicroServiceSensor(
            task_id='wait',
            cws_task_id=f'{group_id}.call',
            timeout=timeout,
        )

        if read:
            tg.read_task = AsyncTechServicePullOperator(
                task_id='read',
                cws_task_id=f'{group_id}.call',
                raise_errors=raise_errors,
                raise_400_errors=raise_400_errors,
                xcom_push=xcom_push,
            )

    tg.call_task >> tg.wait_task

    if transformer:
        tg.transformer_task >> tg.call_task
    if read:
        tg.wait_task >> tg.read_task

    return tg


def NeoRezoServiceAsyncGroup(group_id: str, transformer: t.Callable = None, read: bool = True,
                             op_args: t.Optional[t.Collection[t.Any]] = None,
                             op_kwargs: t.Optional[t.Mapping[str, t.Any]] = None,
                             method: str = 'get', timeout: int = 900,
                             raise_errors: bool = True, raise_400_errors: bool = True,
                             xcom_push: bool = True, trigger_rule=TriggerRule.ALL_SUCCESS,
                             **tech_kwargs):
    """Task group to allow asynchronous call of a TechMicroService.

    The returned value is defined in the task_id : '{group_id}.read'.

    :param group_id: group id.
    :param transformer: python transformer function.
    :param read: do read task.
    :param op_args: python transformer args.
    :param op_kwargs: python transformer kwargs.
    :param method: microservice method called.
    :param raise_errors: raise error on client errors (default True).
    :param raise_400_errors: raise error on client 400 errors (default True).
    :param timeout: asynchronous call timeout.
    :param trigger_rule: trigger rule to be set on first task.
    :param xcom_push:  pushes result in XCom (default True).

    .. versionchanged:: 0.8.4
        Added the ``read`` parameter.
    .. versionchanged:: 0.8.4
        Added the ``trigger_rule`` parameter.
    .. versionchanged:: 0.8.0
        Added the ``transformer`` parameter.
    """
    with CoworksTaskGroup(group_id=group_id) as tg:
        if transformer:
            tg.transformer_task = PythonOperator(
                task_id='transformer',
                python_callable=transformer,
                op_args=op_args,
                op_kwargs=op_kwargs,
                trigger_rule=trigger_rule
            )

            if method.lower() == 'get':
                if 'query_params' in tech_kwargs:
                    msg = "Calling transformer method with already query_params parameter call"
                    tg.transformer_task.log.warning(msg)
                tech_kwargs['query_params'] = tg.transformer_task.output
            else:
                if 'json' in tech_kwargs:
                    tg.transformer_task.log.warning("Calling transformer method with already json parameter call")
                tech_kwargs['json'] = tg.transformer_task.output

        tg.call_task = NeoRezoServiceOperator(
            task_id="call",
            asynchronous=True,
            method=method,
            raise_errors=raise_errors,
            raise_400_errors=raise_400_errors,
            trigger_rule=trigger_rule if not transformer else TriggerRule.ALL_SUCCESS,
            **tech_kwargs,
        )

        tg.wait_task = AsyncTechMicroServiceSensor(
            task_id='wait',
            cws_task_id=f'{group_id}.call',
            timeout=timeout,
        )

        if read:
            tg.read_task = AsyncTechServicePullOperator(
                task_id='read',
                cws_task_id=f'{group_id}.call',
                raise_errors=raise_errors,
                raise_400_errors=raise_400_errors,
                xcom_push=xcom_push,
            )

    tg.call_task >> tg.wait_task

    if transformer:
        tg.transformer_task >> tg.call_task
    if read:
        tg.wait_task >> tg.read_task

    return tg



import tomllib
from pathlib import Path

try:
    pyproject = Path(__file__).resolve().parent.parent / "pyproject.toml"
    with open(pyproject, "rb") as f:
        _META = tomllib.load(f)

    __version__ = _META["project"]["version"]
except FileNotFoundError:
    __version__ = 'deployed'



import base64
import hashlib
import importlib
import inspect
import os
import subprocess
import sys
import sysconfig
import tempfile
import typing as t
from datetime import datetime
from functools import cached_property
from functools import partial
from pathlib import Path
from shutil import ExecError
from shutil import copyfile
from shutil import copytree
from shutil import ignore_patterns
from shutil import make_archive
from subprocess import CompletedProcess

import boto3
import click
from flask.cli import pass_script_info
from flask.cli import with_appcontext
from jinja2 import BaseLoader
from jinja2 import Environment
from jinja2 import PackageLoader
from jinja2 import select_autoescape
from pydantic import BaseModel
from pydantic import ConfigDict
from werkzeug.routing import Rule

from coworks import aws
from coworks.utils import get_cws_annotations
from coworks.utils import load_dotenv
from .command import CwsCommand
from .utils import progressbar
from .utils import show_stage_banner
from .utils import show_terraform_banner

UID_SEP = '_'


class TerraformContext:

    def __init__(self, info, ctx):
        self.app = info.load_app()
        self.ctx = ctx

        # Transform flask app import path into module import path
        if info.app_import_path and '/' in info.app_import_path:
            msg = f"Cannot deploy or destroy a project with handler not on project folder : {info.app_import_path}.\n"
            msg += f"Add option -p {'/'.join(info.app_import_path.split('/')[:-1])} to resolve this."""
            raise ModuleNotFoundError()
        self.app_import_path = info.app_import_path.replace(':', '.') if info.app_import_path else "app.app"


class TerraformResource(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    parent_uid: str | None
    path: str | None
    rules: list[Rule] | None = None

    @cached_property
    def uid(self) -> str:

        if self.is_root:
            return ''

        assert self.path is not None, "TerraformResource Path cannot be None"
        uid = self.path.replace('{', '').replace('}', '')

        if self.parent_is_root:
            return uid

        # avoid too long id for ressources
        parent_uid: str = self.parent_uid if len(self.parent_uid) < 80 else str(id(self.parent_uid))  # type: ignore
        return f"{parent_uid}{UID_SEP}{uid}" if self.path else parent_uid

    @cached_property
    def is_root(self) -> bool:
        return self.path is None

    @cached_property
    def parent_is_root(self) -> bool:
        return self.parent_uid == ''

    @cached_property
    def no_cors_methods(self) -> t.Iterator[set[str] | None]:
        if not self.rules:
            return iter(())
        return (rule.methods for rule in self.rules if getattr(rule, 'cws_no_cors', None))

    def __repr__(self):
        return f"{self.uid}:{self.rules}"


class Terraform:
    """Terraform class to manage local terraform commands."""
    TIMEOUT = 600

    def __init__(self, backend: 'TerraformBackend', terraform_dir, workspace):
        root_context = backend.terraform_context.ctx.find_root()
        self.terraform_context = backend.terraform_context
        self.app_context = backend.terraform_context
        self.bar = backend.bar
        self.terraform_dir = terraform_dir
        self.refresh = backend.terraform_refresh
        self.stage = root_context.params['stage']
        self.workspace = workspace
        self.working_dir = Path(root_context.params.get('project_dir'))

    def init(self):
        self._execute(['init', '-input=false'])

    def apply(self) -> None:
        """Executes terraform apply command."""
        try:
            cmd = ['apply', '-auto-approve', '-parallelism=1']
            if not self.refresh:
                cmd.append('-refresh=false')
            self._execute(cmd)
        except ExecError:
            # on cloud parallelism options is not available
            cmd = ['apply', '-auto-approve']
            if not self.refresh:
                cmd.append('-refresh=false')
            self._execute(cmd)

    def output(self):
        values = self._execute(['output']).stdout
        return values.decode("utf-8").strip()

    @property
    def api_resources(self):
        """Returns the list of flatten path (prev_uid, last, rule)."""
        resources: dict[str, TerraformResource] = {}

        def add_rule(previous: str, path: str, rule_: Rule | None):
            """Add a method rule in a resource."""
            # todo : may use now aws_url_map
            path = None if path is None else path.replace('<', '{').replace('>', '}')
            resource = TerraformResource(parent_uid=previous, path=path)
            if rule_:
                view_function = self.app_context.app.view_functions.get(rule_.endpoint)
                setattr(rule_, 'cws_binary_headers', get_cws_annotations(view_function, '__CWS_BINARY_HEADERS'))
                setattr(rule_, 'cws_no_auth', get_cws_annotations(view_function, '__CWS_NO_AUTH'))
                setattr(rule_, 'cws_no_cors', get_cws_annotations(view_function, '__CWS_NO_CORS'))

            # Creates terraform ressources if it doesn't exist.
            uid = resource.uid
            if uid not in resources:
                resources[uid] = resource

            resource = resources[uid]
            if rule_:
                if resources[uid].rules is None:
                    resources[uid].rules = [rule_]
                else:
                    resources[uid].rules.append(rule_)  # type: ignore[union-attr]
            return uid

        for rule in self.app_context.app.url_map.iter_rules():
            route = rule.rule
            previous_uid = ''
            if route.startswith('/'):
                route = route[1:]
            splited_route = route.split('/')

            # special root case
            if splited_route == ['']:
                add_rule('', '', rule)
                continue

            # creates intermediate resources
            for prev in splited_route[:-1]:
                previous_uid = add_rule(previous_uid, prev, None)

            # set entry keys for last entry
            add_rule(previous_uid, splited_route[-1:][0], rule)

        return resources

    @property
    def logger(self):
        return self.app_context.app.logger

    @property
    def template_loader(self) -> BaseLoader:
        package_name = sys.modules[__name__].__package__
        assert package_name is not None
        return PackageLoader(package_name)

    @property
    def jinja_env(self) -> Environment:
        return Environment(loader=self.template_loader, autoescape=select_autoescape(['html', 'xml']),
                           trim_blocks=True, lstrip_blocks=True)

    def get_context_data(self, **options) -> dict:
        # Microservice context data
        app = self.app_context.app
        tf_key = options.pop('tf_key', None) or app.name.replace('_', '/')
        data = {
            'api_resources': self.api_resources,
            'app': app,
            'app_import_path': self.app_context.app_import_path,
            'debug': app.debug,
            'description': inspect.getdoc(app) or "",
            'environment_variables': load_dotenv(self.stage),
            'ms_name': app.name,
            'now': datetime.now().isoformat(),
            'tf_key': tf_key,
            'stage': self.stage,
            'workspace': self.workspace,
            **options
        }

        # AWS context data
        profile_name = options.get('profile_name')
        if profile_name:
            aws_session = boto3.Session(profile_name=profile_name)
            aws_region = aws_session.region_name
            data['aws_region'] = aws_region
            aws_account = aws_session.client("sts").get_caller_identity()["Account"]
            data['aws_account'] = aws_account

        return data

    def _execute(self, cmd_args: list[str]) -> CompletedProcess:
        self.logger.debug(f"Terraform arguments : ['-chdir={self.terraform_dir} {' '.join(cmd_args)}]")
        p = subprocess.run(["terraform", *cmd_args], capture_output=True, cwd=self.terraform_dir,
                           timeout=self.TIMEOUT)
        if p.returncode != 0:
            msg = p.stderr.decode('utf-8')
            if not msg:
                msg = p.stdout.decode('utf-8')
            raise ExecError(msg)
        return p

    def generate_file(self, template_filename, output_filename, **options) -> None:
        """Generates stage terraform files."""
        template = self.jinja_env.get_template(template_filename)
        with (self.terraform_dir / output_filename).open("w") as f:
            data = self.get_context_data(**options)
            f.write(template.render(**data))
            self.logger.debug(f"Terraform file {self.terraform_dir / output_filename} generated")


class TerraformBackend:
    """Terraform class to manage remote deployements."""

    def __init__(self, terraform_context, bar, **options):
        """The remote terraform class correspond to the terraform command interface for workspaces.
        """
        self.terraform_context = terraform_context
        self.app = terraform_context.app
        self.bar = bar

        # Creates terraform dir if needed
        self.terraform_dir = Path(options['terraform_dir'])
        self.stage = terraform_context.ctx.find_root().params['stage']

        self.terraform_class = Terraform
        self.terraform_refresh = options['terraform_refresh']
        self._api_terraform = self._stage_terraform = None

    @property
    def api_terraform(self):
        if self._api_terraform is None:
            self.app.logger.debug(f"Create common terraform instance using {self.terraform_class}")
            self.terraform_dir.mkdir(exist_ok=True)
            self._api_terraform = self.terraform_class(self, terraform_dir=self.terraform_dir, workspace="common")
        return self._api_terraform

    @property
    def stage_terraform(self):
        if self._stage_terraform is None:
            self.app.logger.debug(f"Create {self.stage} terraform instance using {self.terraform_class}")
            terraform_dir = Path(f"{self.terraform_dir}_{self.stage}")
            terraform_dir.mkdir(exist_ok=True)
            self._stage_terraform = self.terraform_class(self, terraform_dir=terraform_dir, workspace=self.stage)
        return self._stage_terraform

    def process_terraform(self, command_template, terraform_init=True, **options):
        root_command_params = self.terraform_context.ctx.find_root().params

        # Set default options calculated value
        options['key'] = options.get('key') or f"{self.app.name}/archive.zip"

        # Transfert zip file to S3
        self.bar.update(msg="Copy source files on S3")
        b64sha256 = self.copy_sources_to_s3(**options)
        options['source_code_hash'] = b64sha256

        self.bar.update(msg="Generates terraform files")

        # Generates common terraform files
        if terraform_init:
            output_filename = "terraform.tf"
            self.app.logger.debug('Generate terraform global files')
            self.api_terraform.generate_file("terraform.j2", output_filename, **root_command_params, **options)
            self.stage_terraform.generate_file("terraform.j2", output_filename, **root_command_params, **options)

        # Generates terraform files and copy environment variable files in terraform working dir for provisionning
        output_filename = f"{self.app.name}.tech.tf"
        self.app.logger.debug(f'Generate terraform {output_filename} files')
        self.api_terraform.generate_file(command_template, output_filename, **root_command_params, **options)
        self.stage_terraform.generate_file(command_template, output_filename, **root_command_params, **options)

        # Stops process if dry
        if options['dry']:
            self.bar.update(msg="Nothing deployed and destroyed (dry mode)")
            return

        if terraform_init:
            self.app.logger.debug("Init terraform")
            self.api_terraform.init()
            assert self._stage_terraform is not None
            self._stage_terraform.init()

        # Apply to api terraform
        self.bar.update(msg="Create or update API")
        self.api_terraform.apply()
        if options.get('api'):
            return

        # Apply to stage terraform
        self.bar.update(msg=f"Deploy staged lambda ({self.stage})")
        self.stage_terraform.apply()

        self.delete_sources_on_s3(**options)

        # Traces output
        self.bar.terminate()
        click.echo()
        echo_output(self.api_terraform)
        click.secho("🎉 CoWorks Microservice deployed", fg="green")

    def copy_sources_to_s3(self, dry, **options):
        module_name = options.get('module_name') or []
        bucket = options.get('bucket')
        key = options.get('key')
        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_path = Path(tmp_dir)

            # Defines ignore file paternsctx
            ignore = options.get('ignore') or ['.*', 'terraform*']
            if ignore and not isinstance(ignore, list):
                if type(ignore) is tuple:
                    ignore = [*ignore]
                else:
                    ignore = [ignore]
            ignore = [*ignore, '*.pyc', '__pycache__']
            ignore = [*ignore, 'Pipfile*', 'requirements.txt']
            ignore = [*ignore, '*cws.yml', 'env_variables*']
            full_ignore_patterns = partial(ignore_patterns, *ignore)

            # Creates archive
            project_dir = self.terraform_context.ctx.find_root().params.get('project_dir')
            full_project_dir = Path(project_dir).resolve()
            try:
                if tmp_path.relative_to(full_project_dir):
                    msg = f"Cannot deploy a project defined in tmp folder (project dir id {full_project_dir})"
                    raise click.exceptions.UsageError(msg)
            except (Exception,):
                pass

            tmp_sources_path = Path(tmp_path) / 'filtered_dir'
            self.app.logger.debug(f"Copy source file in {tmp_sources_path}")
            copytree(project_dir, tmp_sources_path, ignore=full_ignore_patterns())

            for name in module_name:
                if name.endswith(".py"):
                    file_path = Path(sysconfig.get_path('purelib')) / name
                    copyfile(file_path, tmp_sources_path / name)
                else:
                    mod = importlib.import_module(name)
                    assert mod.__file__ is not None
                    module_path = Path(mod.__file__).resolve().parent
                    copytree(module_path, tmp_sources_path / name, ignore=full_ignore_patterns())
            module_archive = make_archive(str(tmp_path / 'sources'), 'zip', tmp_sources_path)

            # Uploads archive on S3
            with open(module_archive, 'rb') as archive:
                size = int(os.path.getsize(module_archive) / 1000)
                b64sha256 = base64.b64encode(hashlib.sha256(archive.read()).digest()).decode()
                if dry:
                    self.bar.update(msg=f"Nothing copied (dry mode, {size} Kb)")
                else:
                    archive.seek(0)
                    try:
                        profile_name = options.get('profile_name')
                        aws_s3_session = aws.AwsS3Session(profile_name=profile_name)
                        aws_s3_session.client.upload_fileobj(archive, bucket, key)
                    except Exception as e:
                        raise e

                    self.app.logger.debug(msg=f"Successfully uploaded sources at s3://{bucket}/{key}")
                    self.bar.update(msg=f"Sources files copied ({size} Kb)")

            return b64sha256

    def delete_sources_on_s3(self, **options):
        bucket = options.get('bucket')
        key = options.get('key')
        profile_name = options.get('profile_name')
        aws_s3_session = aws.AwsS3Session(profile_name=profile_name)
        aws_s3_session.client.delete_object(Bucket=bucket, Key=key)
        self.bar.update(msg="Sources files deleted on s3")


@click.command("deploy", CwsCommand, short_help="Deploy the CoWorks microservice on AWS Lambda.")
@click.option('--bucket', '-b',
              help="Bucket to upload sources zip file to", required=True)
@click.option('--dry', is_flag=True,
              help="Doesn't perform deploy [Global option only].")
@click.option('--ignore', '-i', multiple=True,
              help="Ignore pattern when copying source to lambda.")
@click.option('--key', '-k',
              help="Sources zip file bucket's name.")
@click.option('--module-name', '-m', multiple=True,
              help="Python module added from current pyenv (module or file.py).")
@click.option('--profile-name', '-pn', required=True,
              help="AWS credential profile.")
@click.option('--binary-types', multiple=True,
              help="Content types defined as binary contents (no encoding).")
@click.option('--json-types', multiple=True,
              help="Add mime types for JSON response.")
@click.option('--layers', '-l', multiple=True, required=True,
              help="Add layer (full arn: aws:lambda:...). Must contains CoWorks at least.")
@click.option('--memory-size', default=128,
              help="Lambda memory size (default 128).")
@click.option('--python', '-p', type=click.Choice(['3.8', '3.9', '3.10', '3.11']), default='3.11',
              help="Python version for the lambda.")
@click.option('--tf-bucket',
              help="Bucket's name if not using cloud terraform.")
@click.option('--tf-key',
              help="Bucket's key if not using cloud terraform (default microservice's name).")
@click.option('--security-groups', multiple=True, default=[],
              help="Security groups to be added [ids].")
@click.option('--subnets', multiple=True, default=[],
              help="Subnets to be added [ids].")
@click.option('--terraform-cloud', '-tc', is_flag=True, default=False,
              help="Use cloud workspaces (default false).")
@click.option('--terraform-dir', '-td', default="terraform",
              help="Terraform files folder (default terraform).")
@click.option('--terraform-organization', '-to',
              help="Terraform organization (needed if using cloud terraform).")
@click.option('--terraform-refresh', '-tr', is_flag=True, default=True,
              help="Forces terraform to refresh the state (default true).")
@click.option('--text-types', multiple=True,
              help="Add mime types for JSON response [at least text/plain, text/html].")
@click.option('--timeout', default=60,
              help="Lambda timeout (default 60s).Only for asynchronous call (API call 30s).")
@click.option('--token-key',
              help="Header token key.")
@click.pass_context
@pass_script_info
@with_appcontext
def deploy_command(info, ctx, **options) -> None:
    """ Deploiement in 2 steps:
        Step 1. Create API and routes integrations
        Step 2. Deploy API and Lambda
    """
    if options.get('terraform_cloud') and not options.get('terraform_organization'):
        raise click.BadParameter('An organization must be defined if using cloud terraform')
    if not options.get('terraform_cloud') and not options.get('tf_bucket'):
        raise click.BadParameter('A bucket must be defined if using terraform on S3 [--tf-bucket option]')

    terraform_context = TerraformContext(info, ctx)
    app = terraform_context.app
    stage = ctx.parent.params['stage']

    app.logger.debug(f"Start deploy command: {options}")
    show_stage_banner(stage)
    cloud = options['terraform_cloud']
    refresh = options['terraform_refresh']
    show_terraform_banner(cloud, refresh)
    with progressbar(label="Deploy microservice", threaded=not app.debug) as bar:  # type: ignore
        terraform_backend_class = options.pop('terraform_class', TerraformBackend)
        app.logger.debug(f"Deploying {app} using {terraform_backend_class}")
        backend = terraform_backend_class(terraform_context, bar, **options)
        backend.process_terraform('deploy.j2', **options)


@click.command("destroy", CwsCommand, short_help="Destroy the CoWorks microservice on AWS Lambda.")
# Zip options (redefined)
@click.option('--bucket', '-b', required=True,
              help="Bucket to upload sources zip file to")
@click.option('--dry', is_flag=True,
              help="Doesn't perform deploy [Global option only].")
@click.option('--key', '-k',
              help="Sources zip file bucket's name.")
@click.option('--profile-name', '-pn', required=True,
              help="AWS credential profile.")
@click.option('--terraform-cloud', is_flag=True, default=False,
              help="Use cloud workspaces (default false).")
@click.option('--terraform-dir', '-td', default="terraform",
              help="Terraform files folder (default terraform).")
@click.option('--terraform-organization', '-to',
              help="Terraform organization needed if using cloud terraform.")
@click.option('--terraform-refresh', '-tr', is_flag=True, default=True,
              help="Forces terraform to refresh the state (default true).")
@click.pass_context
@pass_script_info
@with_appcontext
def destroy_command(info, ctx, **options) -> None:
    """ Destroy by setting counters to 0.
    """
    terraform_context = TerraformContext(info, ctx)
    stage = ctx.parent.params['stage']

    app = terraform_context.app
    app.logger.debug(f"Start destroy command: {options}")
    show_stage_banner(stage)
    cloud = options['terraform_cloud']
    refresh = options['terraform_refresh']
    show_terraform_banner(cloud, refresh)
    with progressbar(label='Destroy microservice', threaded=not app.debug) as bar:  # type: ignore
        terraform_backend_class = options.pop('terraform_class', TerraformBackend)
        app.logger.debug(f'Destroying {app} using {terraform_backend_class}')
        backend = terraform_backend_class(terraform_context, bar, **options)
        backend.process_terraform('destroy.j2', terraform_init=False, **options)
    if not options['dry']:
        click.echo(f"You can now delete the terraform_{options.get('stage')} folder.")


@click.command("deployed", CwsCommand, short_help="Retrieve the microservices deployed for this project.")
@click.option('--terraform-dir', default="terraform",
              help="Terraform folder (default terraform).")
@click.option('--terraform-cloud', is_flag=True, default=False,
              help="Use cloud workspaces (default false).")
@click.option('--terraform-refresh', '-tr', is_flag=True, default=True,
              help="Forces terraform to refresh the state (default true).")
@click.pass_context
@pass_script_info
@with_appcontext
def deployed_command(info, ctx, **options) -> None:
    terraform_context = TerraformContext(info, ctx)
    app = terraform_context.app
    stage = ctx.parent.params['stage']

    app.logger.debug(f"Start destroy command: {options}")
    show_stage_banner(stage)
    cloud = options['terraform_cloud']
    refresh = options['terraform_refresh']
    show_terraform_banner(cloud, refresh)
    with progressbar(label='Retrieving information', threaded=not app.debug) as bar:  # type: ignore
        terraform_backend_class = options.pop('terraform_class', TerraformBackend)
        app.logger.debug(f'Destroying {app} using {terraform_backend_class}')
        backend = terraform_backend_class(terraform_context, bar, **options)
    echo_output(backend.api_terraform)


def echo_output(terraform):
    """Pretty print terraform output.
    """
    rows = terraform.output()
    for row in rows.split('\n'):
        values = row.split('=')
        if len(values) > 1:
            cws_name = values[0].strip()[:-3]  # remove last _id
            api_id = values[1].strip()[1:-1]  # remove quotes
            api_url = f"https://{api_id}.execute-api.eu-west-1.amazonaws.com/"
            click.secho(f"The microservice {cws_name} is deployed at {api_url}", fg='yellow')



import multiprocessing
import os
import time
from unittest import mock

import requests
from flask.cli import ScriptInfo

from coworks.cws.client import client
from coworks.utils import load_dotenv
from tests.coworks.event import get_event
from tests.cws.src.app import EnvTechMS


class TestClass:

    @mock.patch.dict(os.environ, {"FLASK_ENV_FILE": ".env.no", "FLASK_RUN_FROM_CLI": "true"})
    def test_no_env(self, example_dir, empty_aws_context):
        app = EnvTechMS()
        event = get_event('/', 'get')
        with app.cws_client(event, empty_aws_context) as c:
            response = c.get('/', headers={'Authorization': 'token'})

    @mock.patch.dict(os.environ, {"FLASK_ENV_FILE": '.env.dev', "FLASK_RUN_FROM_CLI": "true"})
    def test_run_dev_env(self, example_dir, unused_tcp_port):
        server = multiprocessing.Process(target=run_server, args=(example_dir, unused_tcp_port), daemon=True)
        server.start()
        counter = 1
        time.sleep(counter)
        while not server.is_alive() and counter < 3:
            time.sleep(counter)
            counter += 1
        response = requests.get(f'http://localhost:{unused_tcp_port}/env', headers={'Authorization': "token"})
        assert response.text == "Value of environment variable test is : test dev environment variable."
        server.terminate()

    @mock.patch.dict(os.environ, {"FLASK_ENV_FILE": '.env.v1', "FLASK_RUN_FROM_CLI": "true"})
    def test_run_prod_env(self, example_dir, unused_tcp_port):
        server = multiprocessing.Process(target=run_server, args=(example_dir, unused_tcp_port), daemon=True)
        server.start()
        counter = 1
        time.sleep(counter)
        while not server.is_alive() and counter < 3:
            time.sleep(counter)
            counter += 1
        response = requests.get(f'http://localhost:{unused_tcp_port}/env', headers={'Authorization': "token"})
        assert response.text == "Value of environment variable test is : test prod environment variable."
        server.terminate()

    def test_run_dev_stage(self, example_dir, unused_tcp_port):
        server = multiprocessing.Process(target=run_server_with_stage,
                                         args=(example_dir, unused_tcp_port, "dev"),
                                         daemon=True)
        server.start()
        counter = 1
        time.sleep(counter)
        while not server.is_alive() and counter < 3:
            time.sleep(counter)
            counter += 1
        response = requests.get(f'http://localhost:{unused_tcp_port}/env', headers={'Authorization': "token"})
        assert response.text == "Value of environment variable test is : test dev environment variable."
        server.terminate()

    def test_run_prod_stage(self, example_dir, unused_tcp_port):
        server = multiprocessing.Process(target=run_server_with_stage,
                                         args=(example_dir, unused_tcp_port, "v1"),
                                         daemon=True)
        server.start()
        counter = 1
        time.sleep(counter)
        while not server.is_alive() and counter < 3:
            time.sleep(counter)
            counter += 1
        response = requests.get(f'http://localhost:{unused_tcp_port}/env', headers={'Authorization': "token"})
        assert response.text == "Value of environment variable test is : test prod environment variable."
        server.terminate()

    import pytest
    @pytest.mark.wip
    def test_load_env(self):
        variables = load_dotenv("dev")
        assert True


def run_server(project_dir, port):
    obj = ScriptInfo(create_app=lambda: EnvTechMS(), set_debug_flag=False)
    client.main(['--project-dir', '.', 'run', '--port', port], 'cws', obj=obj, standalone_mode=False)


def run_server_with_stage(project_dir, port, stage):
    @mock.patch.dict(os.environ, {"FLASK_ENV_FILE": f".env.{stage}", "FLASK_RUN_FROM_CLI": "true"})
    def run():
        obj = ScriptInfo(create_app=lambda: EnvTechMS(), set_debug_flag=False)
        client.main(['-p', '.', 'run', '--port', port], 'cws', obj=obj, standalone_mode=False)

    run()



import os
import typing as t
import xmlrpc.client

import requests
from aws_xray_sdk.core import xray_recorder
from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import Field
from pydantic import field_validator
from werkzeug.exceptions import BadRequest
from werkzeug.exceptions import NotFound

from coworks import TechMicroService
from coworks.extension.xray import XRay
from .jsonapi import JsonApiDataMixin
from .jsonapi import JsonApiDict
from .jsonapi.data import CursorPagination


class OdooConfig(BaseModel):
    model_config = ConfigDict(from_attributes=True)

    url: str
    dbname: str
    user: str
    passwd: str
    const: dict[str, t.Any] = Field(default_factory=dict)

    @classmethod
    def from_env_var_prefix(cls, env_var_prefix):
        env_url_var_name = f"{env_var_prefix}_URL"
        env_dbname_var_name = f"{env_var_prefix}_DBNAME"
        env_user_var_name = f"{env_var_prefix}_USER"
        env_passwd_var_name = f"{env_var_prefix}_PASSWD"
        return cls.get_config(env_url_var_name, env_dbname_var_name, env_user_var_name, env_passwd_var_name)

    @classmethod
    def get_config(cls, env_url_var_name: str, env_dbname_var_name: str, env_user_var_name: str,
                   env_passwd_var_name: str):
        url = os.getenv(env_url_var_name)
        if not url:
            raise RuntimeError(f'{env_url_var_name} not defined in environment.')
        dbname = os.getenv(env_dbname_var_name)
        if not dbname:
            raise RuntimeError(f'{env_dbname_var_name} not defined in environment.')
        user = os.getenv(env_user_var_name)
        if not user:
            raise RuntimeError(f'{env_user_var_name} not defined in environment.')
        passwd = os.getenv(env_passwd_var_name)
        if not passwd:
            raise RuntimeError(f'{env_passwd_var_name} not defined in environment.')

        return OdooConfig(url=url, dbname=dbname, user=user, passwd=passwd)


class OdooPagination(CursorPagination):
    """Odoo pagination using limit value and odoo request."""
    max_per_page: int | None = None
    query: t.Any | None = None

    @field_validator("max_per_page")
    def set_max_per_page(cls, max_per_page):
        return max_per_page or 100

    def __iter__(self):
        if self.query:
            return iter(self.query.odoo_execute_kw(self.params))
        raise StopIteration()

    @property
    def params(self):
        assert self.page is not None  # by the validator
        assert self.per_page is not None  # by the validator
        if self.page < 1:
            self.page = 1
        return {
            'limit': self.per_page,
            'offset': (self.page - 1) * self.per_page,
        }


class OdooQuery(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    odoo: "Odoo"
    model: str
    method: str
    order: str | None = None
    fields: list[str] | None = None
    domain: list[tuple[str, str, t.Any]] | None
    bind_key: str | None = None  # key to access the configuration defined in binds

    def paginate(self, *, page=None, per_page=None, max_per_page=None) -> OdooPagination:
        pagination = OdooPagination(query=self, page=page, per_page=per_page, max_per_page=max_per_page, total=0)
        res = self.odoo.odoo_execute_kw(self.model, "search_count", [self.domain], bind_key=self.bind_key)
        pagination.total = res
        return pagination

    def all(self, limit=2) -> list[JsonApiDataMixin]:
        """Mainly used only to get one resource so set limit to 2."""
        params = {
            'limit': limit,
            'offset': 0,
        }
        return self.odoo_execute_kw(params)

    def odoo_execute_kw(self, params):
        res = self.odoo.odoo_execute_kw(self.model, self.method, [self.domain], params, bind_key=self.bind_key)
        return [JsonApiDict(**rec) for rec in res]


class Odoo:
    """Flask's extension for Odoo.
    This extension uses the external API of ODOO.

    At creation: a default configuration must be done or a binding dict oj configurations.
    At execution: if no binding key is provided, the default configuration is used.

    .. versionchanged:: 0.7.3
        ``env_var_prefix`` parameter may be a dict of bind values.
        GraphQL removed.
    """

    def __init__(self, app: TechMicroService | None = None,
                 config: OdooConfig | None = None, binds: dict[str | None, OdooConfig] | None = None):

        """
        :param app: Flask application.
        :param config: default configuration if no bind given in execution.
        :param binds: list of configuration binds.
        """

        self.app = None

        self.binds = binds if binds else {}
        if config:
            self.binds[None] = config

        if app:
            self.init_app(app)

    def init_app(self, app):
        self.app = app

    @XRay.capture(xray_recorder)
    def query(self, model: str, method: str = "search_read", *, fields: list[str] | None = None,
              order: str | None = None, domain: list[tuple[str, str, t.Any]] | None = None,
              bind_key: str | None = None):
        return OdooQuery(odoo=self, model=model, method=method, fields=fields, order=order, domain=domain,
                         bind_key=bind_key)

    @XRay.capture(xray_recorder)
    def kw(self, model: str, method: str = "search_read", *, id: int | None = None, fields: list[str] | None = None,
           order: str | None = None, domain: list[tuple[str, str, t.Any]] | None = None, limit: int = 100,
           page_size: int | None = None, page: int = 0, ensure_one: bool = False, bind_key: str | None = None):
        """Searches with API for records based on the args.
        
        See also: https://www.odoo.com/documentation/14.0/developer/reference/addons/orm.html#odoo.models.Model.search
        @param model: python as a dot separated class name.
        @param method : API method name maybe search_read, search_count
        @param id : id for one element
        @param fields: record fields for result.
        @param order: oder of result.
        @param domain: domain for records.
        @param limit: maximum number of records to return from odoo (default: all).
        @param page_size: pagination done by the microservice.
        @param page: current page searched.
        @param ensure_one: raise error if result is not one (404) and only one (400) object.
        @param bind_key: bind configuration to be used.

        .. versionchanged:: 0.7.3
            Added the ``bind`` parameter.
       """
        if id and domain:
            raise BadRequest("Domain and Id parameters cannot be defined tin same time")
        if id:
            filters = [[('id', '=', id)]]
        else:
            filters = [domain] if domain else [[]]

        params: dict[str, t.Any] = {}
        if order:
            params.update({'order': order})
        if fields:
            params.update({'fields': fields})
        if limit:
            params.update({'limit': limit})
        if page:
            page_size = page_size or limit
            params.update({'offset': (page - 1) * page_size})

        res = self.odoo_execute_kw(model, method, filters, params, bind_key=bind_key)

        if method == 'search_count':
            return res

        if len(res) == 0:
            raise NotFound("No element found.")
        if ensure_one:
            if len(res) != 1:
                raise NotFound("More than one element found and ensure_one parameters was set")
            return res[0]

        return {"ids": [rec['id'] for rec in res], "values": res}

    @XRay.capture(xray_recorder)
    def create(self, model: str, data: list[dict] | None = None, bind_key: str | None = None) -> int:
        """Creates new records for the model.

        See also: https://www.odoo.com/documentation/14.0/developer/reference/addons/orm.html#odoo.models.Model.create
        @param model: python as a dot separated class name.
        @param data: fields, as a list of dictionaries, to initialize and the value to set on them.
        @param bind_key: bind configuration to be used.
        See also:
        https://www.odoo.com/documentation/14.0/developer/reference/addons/orm.html#odoo.models.Model.with_context
        """
        return self.odoo_execute_kw(model, "create", data, bind_key=bind_key)

    @XRay.capture(xray_recorder)
    def write(self, model: str, id: int, data: dict | None = None, bind_key: str | None = None) -> list:
        """Updates one record with the provided values.
        See also: https://www.odoo.com/documentation/14.0/developer/reference/addons/orm.html#odoo.models.Model.write
        @param model: python as a dot separated class name.
        @param id: id of the record.
        @param data: fields to update and the value to set on them.
        @param bind_key: bind configuration to be used.
        """
        return self.odoo_execute_kw(model, "write", [[id], data], bind_key=bind_key)

    @XRay.capture(xray_recorder)
    def delete_(self, model: str, id: int, bind_key: str | None = None) -> list:
        """delete the record.
        See also: https://www.odoo.com/documentation/14.0/developer/reference/addons/orm.html#odoo.models.Model.unlink
        @param model: python as a dot separated class name.
        @param id: id of the record.
        @param bind_key: bind configuration to be used.
        """
        return self.odoo_execute_kw(model, "unlink", [[id]], bind_key=bind_key)

    def get_pdf(self, invoice_id: int, access_token: str, bind_key: str | None = None) -> bytes:
        """Returns the invoice as a PDF invoice.

        @param invoice_id: invoice id.
        @param access_token: access token of this invoice.
        @param bind_key: bind configuration to be used.
        """
        config = self.binds[bind_key]

        url = f"{config.url}/my/invoices/{invoice_id}/?report_type=pdf&download=true&access_token={access_token}"
        resp = requests.get(url)
        if resp.ok:
            return resp.content
        raise NotFound

    def odoo_execute_kw(self, model, method, *args, bind_key: str | None = None, **kwargs):
        """Standard externalm API entries.
        See also: https://www.odoo.com/documentation/15.0/developer/misc/api/odoo.html
        """
        config = self.binds[bind_key]

        if '__uid' not in config.const:
            common = xmlrpc.client.ServerProxy(f'{config.url}/xmlrpc/2/common')
            connected_uid = common.authenticate(config.dbname, config.user, config.passwd, {})
            if not connected_uid:
                raise ConnectionError()
            config.const['__uid'] = connected_uid

        models = xmlrpc.client.ServerProxy(f'{config.url}/xmlrpc/2/object')
        return models.execute_kw(config.dbname, config.const['__uid'], config.passwd, model, method, *args, **kwargs)



import contextlib
import typing as t
from collections import defaultdict
from datetime import datetime

from jsonapi_pydantic.v1_0 import Link
from jsonapi_pydantic.v1_0 import TopLevel
from pydantic.networks import HttpUrl
from sqlalchemy import ColumnOperators
from sqlalchemy import inspect
from sqlalchemy import not_
from sqlalchemy.orm import ColumnProperty
from sqlalchemy.orm import RelationshipProperty
from sqlalchemy.sql import and_
from werkzeug.exceptions import UnprocessableEntity
from werkzeug.local import LocalProxy

from coworks import request
from coworks.utils import nr_url
from coworks.utils import str_to_bool
from .data import JsonApiBaseModel
from .data import JsonApiDataMixin
from .query import Pagination


class FetchingContext:

    def __init__(self, include: str | None = None, fields__: dict[str, str] | None = None,
                 filters__: dict[str, str] | None = None, sort: str | None = None,
                 page__number__: int | None = None, page__size__: int | None = None, page__max__: int | None = None):
        self.include = list(map(str.strip, include.split(','))) if include else []
        self._fields = fields__ if fields__ is not None else {}
        self._sort = list(map(str.strip, sort.split(','))) if sort else []
        self.page = page__number__ or 1
        self.per_page = page__size__ or 100
        self.max_per_page = page__max__ or 100

        # Creates filters dict defined ad jsonapi type as key and attribute expression as value
        self._filters: dict = defaultdict(list)
        filters__ = filters__ if filters__ is not None else {}
        for k, v in filters__.items():
            try:
                json_type, filter = k.rsplit('.', 1)
                self._filters[json_type].append((filter, v))
            except ValueError:
                msg = f"The filter parameter '{k}' must be of the form 'filter[type.value][oper]'"
                raise ValueError(msg)

        self.connection_manager = contextlib.nullcontext()

    def field_names(self, jsonapi_type) -> list[str]:
        """Returns the field's names that must be returned for a specific jsonapi type."""
        if jsonapi_type not in self._fields:
            return []

        fields = self._fields[jsonapi_type]
        if isinstance(fields, list):
            if len(fields) != 1:
                msg = f"Wrong field value '{fields}': multiple fields parameter must be a comma-separated"
                raise UnprocessableEntity(msg)
            field = fields[0]
        else:
            field = fields
        return list(map(str.strip, field.split(',')))

    def pydantic_filters(self, base_model: JsonApiBaseModel):
        _base_model_filters: list[bool] = []
        filter_parameters = fetching_context._filters.get(base_model.jsonapi_type, {})
        for key, value in filter_parameters:
            name, key, oper = self.get_decomposed_key(key)
            if not hasattr(base_model, key):
                msg = f"Wrong '{key}' key for '{base_model.jsonapi_type}' in filters parameters"
                raise UnprocessableEntity(msg)
            column = getattr(base_model, key)

            if oper == 'null':
                if str_to_bool(value[0]):
                    _base_model_filters.append(column is None)
                else:
                    _base_model_filters.append(column is not None)
                continue

            _type = base_model.model_fields.get(key).annotation  # type: ignore[union-attr]
            if _type is bool:
                _base_model_filters.append(base_model_filter(column, oper, str_to_bool(value[0])))
            elif _type is int:
                _base_model_filters.append(base_model_filter(column, oper, int(value[0])))
            elif _type is datetime:
                _base_model_filters.append(
                    base_model_filter(datetime.fromisoformat(column), oper, datetime.fromisoformat(value[0]))
                )
            else:
                _base_model_filters.append(base_model_filter(str(column), oper, value[0]))

        return all(_base_model_filters)

    def sql_filters(self, sql_model: JsonApiDataMixin):
        """Returns the list of filters as a SQLAlchemy filter.

        :param sql_model: the SQLAlchemy model (used to get the SQLAlchemy filter)
        """
        _sql_filters: list[ColumnOperators] = []
        if isinstance(sql_model.jsonapi_type, property):
            jsonapi_type: str = sql_model.jsonapi_type.__get__(sql_model)  # type:ignore[attr-defined]
        else:
            jsonapi_type = sql_model.jsonapi_type
        filter_parameters = self._filters.get(jsonapi_type, {})
        for key, value in filter_parameters:

            # If parameters are defined in body payload they may be not defined as list
            if not isinstance(value, list):
                value = [value]

            name, key, oper = self.get_decomposed_key(key)
            if not hasattr(sql_model, key):
                msg = f"Wrong '{key}' property for sql model '{jsonapi_type}' in filters parameters"
                raise UnprocessableEntity(msg)

            column = getattr(sql_model, key)

            if oper == 'null':
                if len(value) != 1:
                    msg = f"Multiple boolean values '{key}' for null test not allowed"
                    raise UnprocessableEntity(msg)
                exp = column.is_(None) if str_to_bool(value[0]) else not_(column.is_(None))
                _sql_filters.append(exp)

            elif isinstance(column.property, ColumnProperty):
                _type = getattr(column, 'type', None)
                if _type:
                    if _type.python_type is bool:
                        _sql_filters.append(*bool_sql_filter(jsonapi_type, key, column, oper, value))
                    elif _type.python_type is str:
                        _sql_filters.append(*str_sql_filter(jsonapi_type, key, column, oper, value))
                    elif _type.python_type is int:
                        _sql_filters.append(*int_sql_filter(jsonapi_type, key, column, oper, value))
                    elif _type.python_type is datetime:
                        _sql_filters.append(*datetime_sql_filter(jsonapi_type, key, column, oper, value))
                else:
                    _sql_filters.append(column.in_(value))
            elif isinstance(column.property, RelationshipProperty):
                if not isinstance(value, dict):
                    msg = (f"Wrong '{value}' value for sql model '{jsonapi_type}'"
                           " in filters parameters (should be a dict).")
                    raise UnprocessableEntity(msg)
                for k, v in value.items():
                    condition = column.has(**{k: v[0]})
                    if len(v) > 1:
                        for or_v in v[1:]:
                            condition = and_(condition, column.has(**{k: or_v}))
                    _sql_filters.append(condition)

        return _sql_filters

    def sql_order_by(self, sql_model):
        """Returns a SQLAlchemy order from model using fetching order keys.

        :param sql_model: the SQLAlchemy model (used to get the SQLAlchemy order)."""
        insp = inspect(sql_model)
        _sql_order_by = []
        for key in self._sort:
            asc_sort = False
            if key.startswith('-'):
                asc_sort = True
                key = key[1:]

            # sort on relationship
            if '.' in key:
                key, attr = key.split('.', 1)
                if key in insp.all_orm_descriptors:
                    column = insp.all_orm_descriptors[key]
                    raise UnprocessableEntity("Sort on relationship is not implemented")
                else:
                    raise UnprocessableEntity(f"Undefined sort key {key} on model {sql_model}")

            # sort on column attributes
            elif key in insp.column_attrs:
                column = insp.column_attrs[key]
                if asc_sort:
                    column = column.desc()

            _sql_order_by.append(column)
        return _sql_order_by

    def _add_branch(self, tree, vector, value):
        key = vector[0]

        if len(vector) == 1:
            tree[key] = value
        else:
            sub_tree = tree.get(key, {})
            tree[key] = self._add_branch(sub_tree, vector[1:], value)

        return tree

    @staticmethod
    def add_pagination(toplevel: TopLevel, pagination: Pagination):
        if pagination.total > 1:
            links = toplevel.links or {}
            if pagination.has_prev:
                links["prev"] = Link(
                    href=HttpUrl(nr_url(request.path, {"page[number]": pagination.prev_num}, merge_query=True))
                )
            if pagination.has_next:
                links["next"] = Link(
                    href=HttpUrl(nr_url(request.path, {"page[number]": pagination.next_num}, merge_query=True))
                )
            toplevel.links = links

        meta = toplevel.meta or {}
        meta["count"] = pagination.total
        meta["pagination"] = {
            "page": pagination.page,
            "pages": pagination.pages,
            "per_page": pagination.per_page,
        }
        toplevel.meta = meta

    def get_decomposed_key(self, key) -> tuple[str | None, str, str | None]:
        name = oper = None

        # filter operator
        # idea from https://discuss.jsonapi.org/t/share-propose-a-filtering-strategy/257
        if "____" in key:
            key, oper = key.split('____', 1)

        # if the key is named (for allowing composition of filters)
        if ':' in key:
            name, key = key.split(':', 1)

        return name, key, oper


def create_fetching_context_proxy(include: str | None = None, fields__: dict | None = None,
                                  filters__: dict | None = None, sort: str | None = None,
                                  page__number__: int | None = None, page__size__: int | None = None,
                                  page__max__: int | None = None):
    context = FetchingContext(include, fields__, filters__, sort, page__number__, page__size__, page__max__)
    setattr(request, 'fetching_context', context)


fetching_context = t.cast(FetchingContext,
                          LocalProxy(lambda: getattr(request, 'fetching_context', 'Not in JsonApi context')))


def base_model_filter(column, oper, value) -> bool:
    """String filter."""
    oper = oper or 'eq'
    if oper == 'eq':
        return column == value
    if oper == 'neq':
        return column != value
    if oper == 'contains':
        return value in column
    msg = f"Undefined operator '{oper}' for string value"
    raise UnprocessableEntity(msg)


def bool_sql_filter(jsonapi_type, key, column, oper, value) -> list[ColumnOperators]:
    """Boolean filter."""
    if len(value) != 1:
        msg = f"Multiple boolean values '{key}' property on model '{jsonapi_type}' is not allowed"
        raise UnprocessableEntity(msg)
    return [column == str_to_bool(value[0])]


def str_sql_filter(jsonapi_type, key, column, oper, value) -> list[ColumnOperators]:
    """String filter."""
    oper = oper or 'eq'
    if oper == 'eq':
        return [column.in_(value)]
    if oper == 'ilike':
        return [column.ilike(str(v)) for v in value]
    if oper == 'contains':
        return [column.contains(str(v)) for v in value]
    msg = f"Undefined operator '{oper}' for string value"
    raise UnprocessableEntity(msg)


def int_sql_filter(jsonapi_type, key, column, oper, value) -> list[ColumnOperators]:
    """Datetime filter."""
    oper = oper or 'eq'
    if oper not in ('eq', 'neq', 'ge', 'gt', 'le', 'lt'):
        msg = f"Undefined operator '{oper}' for integer value"
        raise UnprocessableEntity(msg)
    return [sort_operator(column, oper, int(v)) for v in value]


def datetime_sql_filter(jsonapi_type, key, column, oper, value) -> list[ColumnOperators]:
    """Datetime filter."""
    oper = oper or 'eq'
    if oper not in ('eq', 'neq', 'ge', 'gt', 'le', 'lt'):
        msg = f"Undefined operator '{oper}' for datetime value"
        raise UnprocessableEntity(msg)
    return [sort_operator(column, oper, datetime.fromisoformat(v)) for v in value]


def sort_operator(column: t.Any, oper, value) -> t.Any:
    if oper == 'eq':
        return column == value
    if oper == 'neq':
        return column != value
    if oper == 'ge':
        return column >= value
    if oper == 'gt':
        return column > value
    if oper == 'le':
        return column <= value
    if oper == 'lt':
        return column < value
    msg = f"Undefined operator '{oper}' in sort_operator"
    raise UnprocessableEntity(msg)
